/*
 * cethernet.c
 * SeaMAX for Linux
 *
 * This code implements the Ethernet discovery and configuration
 * portion of the SeaMAX for Linux library.
 *
 * Sealevel and SeaMAX are registered trademarks of Sealevel Systems
 * Incorporated.
 *
 *  2008-2017 Sealevel Systems, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the Lesser GNU General Public License
 * as published by the Free Software Foundation; either version
 * 3 of the License, or (at your option) any later version.
 * LGPL v3
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <net/if.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <linux/if_ether.h>
#include <net/ethernet.h>
#include <netpacket/packet.h>

#include "cethernet.h"

#define CETH_MAX_INF 5  //Maximum number of interfaces to cycle through.

// Global variables.
int sock = -1;
int bsocks[CETH_MAX_INF];
int bsockcount = 0;

// ============================================================================
// ============================================================================
// The following are internal helper functions, that the user should not use.
// ============================================================================
// ============================================================================

//  --------------------------------------------------------------------------
// ( Private function to insert a ceth_device into an allocated list.         )
//  --------------------------------------------------------------------------
int addDevice(unsigned long IP, unsigned char *buff, int len, ceth_device *list)
{
	ceth_device *new = list;

	// This prevents duplicate devices.
	while (new != NULL && !(new->ip_address.i == 0 || new->ip_address.i == IP))
		new = new->next;

	// Overflow
	if (new == NULL) return 0;

	// Device specific information.
	int rc = 0;
	if (new->ip_address.i == 0)
	{
		new->type = SeaIO_Ethernet;
		new->ip_address.i = IP;
		rc = 1;
	}

	if (buff[3] == 0xF7 && len >= 30)
	{
		// save mac address
		memcpy(new->mac_address.c, &buff[24], 6);
	}
	else if (buff[3] == 0xD0 && len >= 124)
	{
		// If the returned IP isn't 0, then it's not DHCP enabled.
		new->dhcp_enabled = ((buff[4] | buff[5] | buff[6] | buff[7]) == 0);

		// Gateway address
		if (new->dhcp_enabled) memset(new->gateway.c, 0x00, 4);
		else memcpy(new->gateway.c, &buff[16], 4);

		// Netmask
		new->net_mask.i = 0;
		new->net_mask.i = (~new->net_mask.i) >> buff[10];
		if (new->dhcp_enabled || buff[10] == 0)
			new->net_mask.i = 0x00FFFFFF;

		// First half of DHCP name
		memcpy(new->name, &buff[116], 8);
	}
	else if (buff[3] == 0xD3 && len >= 130)
	{
		// Second half of DHCP name
		memcpy(&new->name[8], &buff[106], 8);
	}

	// We'll get the MAC address, after we've found all the devices.
	return rc;
}

//  --------------------------------------------------------------------------
// ( Private function to return the length of an allocated space.  Chain!     )
//  --------------------------------------------------------------------------
int listLength(ceth_device *list)
{
	int i = 0;
	ceth_device *cursor;

	for (cursor = list; cursor != NULL; cursor = cursor->next)
		i++;

	return i;
}

//  --------------------------------------------------------------------------
// ( Private function to clean any data in the current list.                  )
//  --------------------------------------------------------------------------
void clearList(ceth_device *list)
{
	ceth_device *cursor;

	for (cursor = list; cursor != NULL; cursor = cursor->next)
	{
		ceth_device *temp = cursor->next;
		bzero(cursor, sizeof(ceth_device));
		cursor->next = temp;
	}
}

//  --------------------------------------------------------------------------
// ( Private function to setup broadcast sockets bound to all interfaces.     )
//  --------------------------------------------------------------------------
int udpBroadcastInit()
{
	// Get a list of interfaces
	int dummy = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (dummy < 0) return -ENOTCONN;

	struct ifreq interface[CETH_MAX_INF];
	bzero(interface, sizeof(interface));

	struct ifconf iflist;
	iflist.ifc_len = sizeof(interface);
	iflist.ifc_req = interface;
	if (ioctl(dummy, SIOCGIFCONF, &iflist) < 0) return -ENOTCONN;
	close(dummy);

	// Clear broadcast socket list if it already exists
	bsockcount = 0;
	for (int i = 0; i < CETH_MAX_INF; i++)
	{
		close(bsocks[i]);
		bsocks[i] = 0;
	}

	// Create and bind a new socket for each valid interface
	for (int i = 0; i < CETH_MAX_INF; i++)
	{
		// This needs to be a valid interface (IPv4 only).
		unsigned int *IP = (unsigned int*) &interface[i].ifr_addr.sa_data[2];
		if (interface[i].ifr_addr.sa_family == AF_INET && *IP != 0 && *IP != 0x0100007F)
		{
			// create new socket
			bsocks[bsockcount] = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
			if (bsocks[bsockcount] < 0) return -ENOTCONN;

			// bind the socket to the interface (this operation requires elevated privileges)
			int status = setsockopt(bsocks[bsockcount], SOL_SOCKET, SO_BINDTODEVICE, interface[i].ifr_name, IFNAMSIZ);
			if (status < 0) return -EPERM;

			// enable broadcasting on the socket
			unsigned int opt = 1;
			status = setsockopt(bsocks[bsockcount], SOL_SOCKET, SO_BROADCAST, &opt, sizeof(opt));
			if (status < 0) return -ENOTCONN;

			// set a receive timeout
			struct timeval delay;
			delay.tv_sec = 0;
			delay.tv_usec = 1000 * 100;  //100 ms
			status = setsockopt(bsocks[bsockcount], SOL_SOCKET, SO_RCVTIMEO, &delay, sizeof(delay));
			if (status < 0) return -ENOTCONN;

			bsockcount++;
		}
	}

	return 0;
}

//  --------------------------------------------------------------------------
// ( Private function to make a udp broadcasts across all available interfaes.)
//  --------------------------------------------------------------------------
int udpBroadcast(unsigned char *mess, int len, unsigned int port)
{
	// Check arguments.
	if (mess == NULL) 	return -EINVAL;
	if (len <= 0) 		return -EINVAL;
	if (port <= 0) 		return -EINVAL;

	// send the message on each broadcast socket
	for (int i = 0; i < bsockcount; i++)
	{
		// destination information struct.  255.255.255.255 = broadcast
		struct sockaddr_in dest;
		dest.sin_family = AF_INET;
		dest.sin_addr.s_addr = inet_addr("255.255.255.255");
		dest.sin_port = htons(port);

		sendto(bsocks[i], mess, len, 0, (struct sockaddr*) &dest, sizeof(dest));
	}

	return len;
}

//  --------------------------------------------------------------------------
// ( Private function to recv the first available udp response from the       )
// ( broadcast socket list.                                                   )
//  --------------------------------------------------------------------------
int udpBroadcastRecv(unsigned char *mess, int len, unsigned long *IP, unsigned int port)
{
	// Check arguments.
	if (mess == NULL) 	return -EINVAL;
	if (len <= 0)		return -EINVAL;
	if (IP == NULL)		return -EINVAL;
	if (port <= 0)		return -EINVAL;

	// Get a response
	for (int i = 0; i < bsockcount; i++)
	{
		struct sockaddr_in from;
		unsigned int size = sizeof(struct sockaddr);

		bzero(mess, len);
		int status = recvfrom(bsocks[i], mess, len, 0, (struct sockaddr*) &from, &size);
		if (status <= 0) continue;

		// Return the IP address
		*IP = from.sin_addr.s_addr;
		return status;
	}

	return 0;
}

//  --------------------------------------------------------------------------
// ( Private function to make a udp request.  dest=255.255.255.255 bcast      )
//  --------------------------------------------------------------------------
int udpSend(unsigned char *mess, int len, char *strdest, unsigned int port)
{
	int status;
	unsigned int opt = 1;
	struct sockaddr_in dest;

	// Check arguments.
	if (mess == NULL) 	return -EINVAL;
	if (len <= 0) 		return -EINVAL;
	if (strdest == NULL) 	return -EINVAL;
	if (port <= 0) 		return -EINVAL;

	// Make sure a socket is open.
	if (sock < 0) sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock < 0) return -ENOTCONN;

	// Allow broadcasting.
	status = setsockopt(sock, SOL_SOCKET, SO_BROADCAST, &opt, sizeof(opt));
	if (status < 0) return -ENOTCONN;

	// Destination information struct.  255.255.255.255 = broadcast.
	dest.sin_family = AF_INET;
	dest.sin_addr.s_addr = inet_addr(strdest);
	dest.sin_port = htons(port);
	
	// Send the supplied buffer.
	return sendto(sock,mess,len,0,(struct sockaddr*) &dest,sizeof(dest));
}

//  --------------------------------------------------------------------------
// ( Private function to receive udp responses.  IP is returned.              )
//  --------------------------------------------------------------------------
int udpRecv(unsigned char *mess, int len, unsigned long *IP, unsigned int port)
{
	int status;
	struct sockaddr_in from;
	struct timeval delay;
	unsigned int size = sizeof(struct sockaddr);

	// Check arguments.
	if (mess == NULL) 	return -EINVAL;
	if (len <= 0)		return -EINVAL;
	if (IP == NULL)		return -EINVAL;
	if (port <= 0)		return -EINVAL;

	// Clear any garbage in the buffer.
	bzero(mess, len);

	// Make sure a socket is open.
	if (sock < 0) return -ENOTCONN;

	// Set a receive timeout.
	delay.tv_sec = 0;
	delay.tv_usec = 1000 * 50;  //50 ms
	status = setsockopt(sock,SOL_SOCKET,SO_RCVTIMEO,&delay,sizeof(delay));
	if (status < 0) return -ENOTCONN;

	// Get a response
	status = recvfrom(sock, mess, len, 0, (struct sockaddr*) &from, &size);

	// Make sure the response is on the right port.
	if (from.sin_port != htons(port))
	{
		bzero(mess, len);  //Whatever we got, we don't want it.
		return -EBADMSG;   //Return an error.
	}

	// Return the IP address, will not be human readable.
	*IP = from.sin_addr.s_addr;

	return status;
}

//  --------------------------------------------------------------------------
// ( Private function to make the requests and fill in the received data.     )
//  --------------------------------------------------------------------------
int findDevices(ceth_device_type type, int number, ceth_device *list)
{
	// Argument check
	if (type == SeaLink_Dev) return 0;  //Functionality unsupported.
	if (number <= 0) 	 return 0;  //Find negative devices?
	if (list == NULL) 	 return 0;  //Where do we store info?

	// Store as many devices as we can.
	if (number > listLength(list)) number = listLength(list);

	// To request firmware version and MAC a message of 0x"00 01 00 F6" is used
	unsigned char buffer[256];
	bzero(buffer, sizeof(buffer));
	buffer[1] = 0x01;
	buffer[3] = 0xF6;

	// This function will setup a list of broadcast sockets for each interface
	int status = udpBroadcastInit();
	if (status < 0) return status;

	// Send the message on each broadcast socket to port 0x77FE
	status = udpBroadcast(buffer, 4, 0x77FE);
	if (status < 0) return status;

	// Clear any data we found on a previous scan.
	clearList(list);

	usleep(1000 * 50);

	// Now get all the responses we can...
	int devs = 0;
	status = 1;
	while (status > 0)
	{
		// Response is from this IP.
		unsigned long IP;

		status = udpBroadcastRecv(buffer, sizeof(buffer), &IP, 0x77FE);
		if (devs < number) devs += addDevice(IP, buffer, status, list);
	}

	// To request setup record zero a message of 0x"00 01 00 E0" is used
	bzero(buffer, sizeof(buffer));
	buffer[1] = 0x01;
	buffer[3] = 0xE0;

	// Send the message on each broadcast socket to port 0x77FE
	status = udpBroadcast(buffer, 4, 0x77FE);
	if (status < 0) return status;

	usleep(1000 * 50);

	// Now get all the responses we can...
	status = 1;
	while (status > 0)
	{
		// Response is from this IP.
		unsigned long IP;

		status = udpBroadcastRecv(buffer, sizeof(buffer), &IP, 0x77FE);
		if (devs < number) devs += addDevice(IP, buffer, status, list);
	}

	// To request setup record three a message of 0x"00 01 00 E3" is used
	bzero(buffer, sizeof(buffer));
	buffer[1] = 0x01;
	buffer[3] = 0xE3;

	// Send the message on each broadcast socket to port 0x77FE
	status = udpBroadcast(buffer, 4, 0x77FE);
	if (status < 0) return status;

	usleep(1000 * 50);

	// Now get all the responses we can...
	status = 1;
	while (status > 0)
	{
		// Response is from this IP.
		unsigned long IP;

		status = udpBroadcastRecv(buffer, sizeof(buffer), &IP, 0x77FE);
		if (devs < number) devs += addDevice(IP, buffer, status, list);
	}

	return devs;
}


// ============================================================================
// ============================================================================
// CEthernet API calls (public)
// ============================================================================
// ============================================================================


// ----------------------------------------------------------------------------
/// \ingroup group_cethernet_fun
/// \brief Device struct allocation.
/// This function will allocate space for a given number of ceth_devices.  You
/// MUST Alloc() space for at least as many devices as you desire to find.
///
/// \param[in] number The number of devices you wish to allocate space for.
///
/// \return *ceth_device  Pointer to the first allocated device.
/// \retval -ENOMEM       Low memory.
// ----------------------------------------------------------------------------
ceth_device *CEthernet_Alloc(int number)
{
	// Pointer to the current chunk of memory being used.
	ceth_device *list;

	// This conditional will not only prevent negative sizes, it will
	// stop the recursion.
	if (number <= 0) return NULL;

	// Allocate space for the current device, if there is no space... ERR
	list = malloc(sizeof(ceth_device));
	if (list == NULL) return (ceth_device*) -ENOMEM;

	// Clear the data of the current device struct
	bzero(list, sizeof(ceth_device));

	// Now recurse, setting this device's next pointer as the result.
	list->next = CEthernet_Alloc(number - 1);
	if (list->next < 0)
	{
		free(list);
		return (ceth_device*) -ENOMEM;
	}

	// Return the pointer to this device, which chains on subsequent devices
	return list;
}


// ----------------------------------------------------------------------------
/// \ingroup group_cethernet_fun
/// \brief Device struct de-allocation.
/// You MUST free the memory allocated by Alloc() when you are done.
///
/// \param[in] *list   Pointer to the list of devices previously allocated.
///
/// \return void
// ----------------------------------------------------------------------------
void CEthernet_Free(ceth_device *list)
{
	// Recursive break point, also do nothing if called on empty list
	if (list == NULL) return;

	// Recursively free any and all children.
	CEthernet_Free(list->next);

	// Clear the information for the current struct
	bzero(list, sizeof(ceth_device));

	// Now release the memory being used for the current struct.
	free(list);
}

// ----------------------------------------------------------------------------
/// \ingroup group_cethernet_fun
/// \brief This function broadcasts a request for all devices to identify 
/// themselves.
/// A series of udp broadcast packets are sent out on a special port.  Sealevel
/// devices listening on this port will respond with device configuration
/// information including IP, Netmask, Gateway, Name, and MAC address.
/// NOTE: In order to support multiple ethernet interfaces, this function must
/// use the SO_BINDTODEVICE socket option.  This option requires elevated
/// privileges.
///
/// \param[in] type    Search for SeaLink or SeaIO devices.
/// \param[in] num     The maximum number of devices to find.  Size of list.
/// \param[out] *list  Information about found devices will be stored here.
///
/// \return int        Error code.
/// \retval >=0        Number of devices found.
/// \retval -EPERM     Insufficent privileges.
/// \retval -EINVAL    Incorrect parameters.
/// \retval -ENOTCONN  Error establishing socket connection.
/// \retval -EBADMSG   Received a response from an unknown port.
// ----------------------------------------------------------------------------
int CEthernet_find_devices(ceth_device_type type, int num, ceth_device *list)
{
	int found = findDevices(type, num, list);
	return found;
}

// ----------------------------------------------------------------------------
/// \ingroup group_cethernet_fun
/// \brief Set a device's Ethernet configuration.
/// This function can be used to change a device's DHCP name, IP, NetMask,
/// and Gateway.  You may also use this function to set a device to DHCP, IF
/// it is still on your local subnet.  If the device is no longer on your
/// subnet, you may use the recover_module() function to re-enable DHCP mode.
/// When you use the command SetIPAddress, you must supply IP, NetMask, and
/// Gateway ceth_ip_address structs.  Error checking will determine if you
/// supply invalid IP and NetMask.  ie a broadcast/home IP or non-standard 
/// netmask.
///
/// You have the option of OR'ing an IP command with the SetName command.
/// For example you may SetDHCP while also renaming a device like so:
/// set_information(ceth_device*, SeaIO_Ethernet, SetDHCP & SetName, "Name");
/// If you SetIPAddress AND SetName, you MUST provide the name as the LAST
/// parameter.
///
/// \param[in] *device  Pointer to the device you would like to modify.
/// \param[in] command  The type of information you desire to set.
/// \param     ...      The data you wish to set.  Depends on command.
///
/// \return int        Error code.
/// \retval >=0        Number of devices found.
/// \retval -EINVAL    Incorrect parameters.
/// \retval -EPERM     Communication error.
// ----------------------------------------------------------------------------
int CEthernet_set_information(ceth_device *device, ceth_set_types command, ...)
{
	va_list ap;
	struct in_addr IP;
	unsigned char buffer[256];
	unsigned char buff2[20];
	char *name = NULL;
	ceth_ip_addr ip, nm, gw;
	ip.i = nm.i = gw.i = 0;

	// Check parameters for consistency.
	if (device == NULL) return -EINVAL;
	if (command < 0 || command == 3 || command >= 11) return -EINVAL;

	// Enable access to other parameters.
	va_start(ap, command);

	if (command & SetIPAddress)
	{
		int i;
		ip.i = va_arg(ap, unsigned int);
		nm.i = va_arg(ap, unsigned int);
		gw.i = va_arg(ap, unsigned int);

		// Check for ip/nm validity.
		if (ip.i == 0 || ip.i == 0xFFFFFFFF) return -EINVAL;
		if (nm.i == 0 || nm.i == 0xFFFFFFFF) return -EINVAL;

		for (i = 0; i < 32; i++)
			if (((nm.i >> i) & 1) == 0) break;
		for (     ; i < 32; i++)
			if (((nm.i >> i) & 1) == 1) return -EINVAL;

		if (ip.i == (nm.i & ip.i)) 	return -EINVAL;
		if (ip.i == (~nm.i | ip.i)) 	return -EINVAL;
	}
	if (command & SetName)
		name = va_arg(ap, char*);

	// Done getting other parms
	va_end(ap);

	// Now request the current setup record 0.
	bzero(buffer, sizeof(buffer));
	buffer[3] = 0xE0;
	IP.s_addr = device->ip_address.i;

	udpSend(buffer, sizeof(buffer), inet_ntoa(IP), 0x77FE);
	if(udpRecv(buffer,sizeof(buffer),(unsigned long*) &IP.s_addr,0x77FE)<0)
		return -EPERM;

	// Finally make our changes and apply them
	buffer[3] = 0xC0;
	if (command & SetIPAddress || command & SetDHCP)
	{
		int i;
		// IP and Gateway addresses
		for (i = 0; i < 4; i++)
		{
			buffer[4+i] = ip.c[i];
			buffer[16+i] = gw.c[i];
		}
		// Determine number of host bits
		for (i = 0; i < 32; i++)
			if (((nm.i >> i) & 1) == 0) break;
		buffer[10] = 32 - i;
	}
	if (command & SetName)
	{
		int i;
		for (i = 0; i < 8; i++)
			if (i < strlen(name)) buffer[116+i] = name[i];
			else buffer[116+i] = 0x00;
	}

	IP.s_addr = device->ip_address.i;
	udpSend(buffer, sizeof(buffer), inet_ntoa(IP), 0x77FE);

	if(udpRecv(buff2,sizeof(buff2),(unsigned long*) &IP.s_addr,0x77FE) < 0)
		return -EPERM;
	if (buff2[3] != 0xB0) return -EPERM;

	// Update ceth_device struct settings.
	if (command & SetIPAddress)
		device->ip_address.i = ip.i;
	else
		device->ip_address.i = IP.s_addr;

	addDevice(device->ip_address.i, buffer, sizeof(buffer), device);

	return 0;
}

// ----------------------------------------------------------------------------
/// \ingroup group_cethernet_fun
/// \brief Use this function to recover a device not on your subnet.
/// This function is designed to re-enable DHCP on a device that has been
/// removed from your subnet by a SetIPAddress command.  Since you can no
/// longer send commands to a device that is no longer on your subnet, it
/// is helpful to be able to re-enable DHCP so that you may once again.
///  NOTE:  This function requires use of the MAC address and is therefore
///           a privileged (RAW Socket) function.  You may alternatively hard
///           code the device's MAC address into the ceth_device struct provided
///           to this function.
///
/// \param[in] *device  Pointer to the device you would like to recover.
///
/// \return int        Error code.
/// \retval 0          Success.
/// \retval -1         No device specified.
/// \retval -2         MAC address unknown.
/// \retval -3         DHCP is already enabled.
/// \retval -4         Failed to find the device.  Recovery failed.
// ----------------------------------------------------------------------------
int CEthernet_recover_module(ceth_device *device)
{
	struct hostent *host;
	unsigned char buffer[20] = "    IP-SETUP";
	if (device == NULL) return -1;

	if (device->mac_address.c[0] == 0 && device->mac_address.c[1] == 0
	 && device->mac_address.c[2] == 0 && device->mac_address.c[3] == 0
	 && device->mac_address.c[4] == 0 && device->mac_address.c[5] == 0)
		return -2;
	
	if (device->dhcp_enabled) return -3;

	// Format broadcast dhcp enable request.
	bzero(buffer, 4);
	buffer[3] = 0xFC;
	bzero(&buffer[12], 8);
	memcpy(&buffer[14], &device->mac_address.c[4], 2);

	// Send the request.
	udpSend(buffer, sizeof(buffer), "255.255.255.255", 0x77FE);

	// Now the device should be rebooting and shaking hands.
	// So we need to wait a little while, then look for him.
	usleep(1000*50);  //50ms pause
	host = gethostbyname(device->name);
	if (host == NULL) return -4;
	
	device->ip_address.i = *((unsigned long*) host->h_addr_list[0]);
	device->gateway.i = 0x00000000;   //Device still keeps a gateway
	device->net_mask.i = 0x00FFFFFF;  //DHCP standard... (silly)

	return 0;
}
