/*
 * cethernet.cpp
 * SeaMAX for Linux
 *
 * This code implements a C++ wrapper class for the Ethernet discovery
 * and configuration portion of the SeaMAX for Linux library.
 *
 * Sealevel and SeaMAX are registered trademarks of Sealevel Systems
 * Incorporated.
 *
 *  2008-2017 Sealevel Systems, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the Lesser GNU General Public License
 * as published by the Free Software Foundation; either version
 * 3 of the License, or (at your option) any later version.
 * LGPL v3
 */

#include <stdlib.h>
#include <stdarg.h>

#include "cethernet.h"

// ----------------------------------------------------------------------------
/// \ingroup group_cethernet_oop
/// \brief Device struct allocation.
/// This function will allocate space for a given number of ceth_devices.  You
/// MUST Alloc() space for at least as many devices as you desire to find.
///
/// \param[in] number The number of devices you wish to allocate space for.
///
/// \return *ceth_device  Pointer to the first allocated device.
/// \retval -ENOMEM       Low memory.
// ----------------------------------------------------------------------------
ceth_device *CCEthernet::Alloc(int number)
{
	return CEthernet_Alloc(number);
}

// ----------------------------------------------------------------------------
/// \ingroup group_cethernet_oop
/// \brief Device struct de-allocation.
/// You MUST free the memory allocated by Alloc() when you are done.
///
/// \param[in] *list   Pointer to the list of devices previously allocated.
///
/// \return void
// ----------------------------------------------------------------------------
void CCEthernet::Free(ceth_device *list)
{
	return CEthernet_Free(list);
}

// ----------------------------------------------------------------------------
/// \ingroup group_cethernet_oop
/// \brief This function broadcasts a request for all devices to identify 
/// themselves.
/// A series of udp broadcast packets are sent out on a special port.  Sealevel
/// devices listening on this port will respond with device configuration
/// information including IP, Netmask, Gateway, Name, and MAC address.
/// NOTE: In order to support multiple ethernet interfaces, this function must
/// use the SO_BINDTODEVICE socket option.  This option requires elevated
/// privileges.
///
/// \param[in] type    Search for SeaLink or SeaIO devices.
/// \param[in] num     The maximum number of devices to find.  Size of list.
/// \param[out] *list  Information about found devices will be stored here.
///
/// \return int        Error code.
/// \retval >=0        Number of devices found.
/// \retval -EPERM     Insufficent privileges.
/// \retval -EINVAL    Incorrect parameters.
/// \retval -ENOTCONN  Error establishing socket connection.
/// \retval -EBADMSG   Received a response from an unknown port.
// ----------------------------------------------------------------------------
int CCEthernet::find_devices(ceth_device_type type, int num, ceth_device *list)
{
	return CEthernet_find_devices(type, num, list);
}

// ----------------------------------------------------------------------------
/// \ingroup group_cethernet_oop
/// \brief Set a device's Ethernet configuration.
/// This function can be used to change a device's DHCP name, IP, NetMask,
/// and Gateway.  You may also use this function to set a device to DHCP, IF
/// it is still on your local subnet.  If the device is no longer on your
/// subnet, you may use the recover_module() function to re-enable DHCP mode.
/// When you use the command SetIPAddress, you must supply IP, NetMask, and
/// Gateway ceth_ip_address structs.  Error checking will determine if you
/// supply invalid IP and NetMask.  ie a broadcast/home IP or nonstandard 
/// netmask.
///
/// You have the option of OR'ing an IP command with the SetName command.
/// For example you may SetDHCP while also renaming a device like so:
/// set_information(ceth_device*, SeaIO_Ethernet, SetDHCP & SetName, "Name");
/// If you SetIPAddress AND SetName, you MUST provide the name as the LAST
/// parameter.
///
/// \param[in] *device  Pointer to the device you would like to modify.
/// \param[in] command  The type of information you desire to set.
/// \param     ...      The data you wish to set.  Depends on command.
///
/// \return int        Error code.
/// \retval >=0        Number of devices found.
/// \retval -EINVAL    Incorrect parameters.
/// \retval -EPERM     Communication error.
// ----------------------------------------------------------------------------
int CCEthernet::set_information(ceth_device *device, ceth_set_types command, 
									   ...)
{
	va_list ap;
	char *name = NULL;
	ceth_ip_addr ip, nm, gw;

	// Enable access to other parameters.
	va_start(ap, command);

	if (command & SetIPAddress)
	{
		ip.i = va_arg(ap, unsigned int);
		nm.i = va_arg(ap, unsigned int);
		gw.i = va_arg(ap, unsigned int);
	}
	if (command & SetName)
		name = va_arg(ap, char*);

	// Done getting other parms
	va_end(ap);

	// Trick the mulitple arguments into working correctly.
	if (command == SetName || command == (SetDHCP & SetName))
		return CEthernet_set_information(device, command, name);
	else
		return CEthernet_set_information(device, command, ip, nm, gw,
			       						name);

}

// ----------------------------------------------------------------------------
/// \ingroup group_cethernet_oop
/// \brief Use this function to recover a device not on your subnet.
/// This function is designed to re-enable DHCP on a device that has been
/// removed from your subnet by a SetIPAddress command.  Since you can no
/// longer send commands to a device that is no longer on your subnet, it
/// is helpful to be able to re-enable DHCP so that you may once again.
///  NOTE:  This function requires use of the MAC address and is therefore
///           a privileged (RAW Socket) function.  You may alternatively hard
///           code the device's MAC address into the ceth_device struct provided
///           to this function.
///
/// \param[in] *device  Pointer to the device you would like to recover.
///
/// \return int        Error code.
/// \retval 0          Success.
/// \retval -1         No device specified.
/// \retval -2         MAC address unknown.
/// \retval -3         DHCP is already enabled.
/// \retval -4         Failed to find the device.  Recovery failed.
// ----------------------------------------------------------------------------
int CCEthernet::recover_module(ceth_device *device)
{
	return CEthernet_recover_module(device);
}
