/*
 * TestIt.cpp
 * SeaMAX for Linux Sealevel Systems Diagnostic
 *
 * This code implements the linssd utility.  This utility is a command line
 * or batch operation tool to aid in simple configuration, test, and demonstration 
 * of SeaMAX supported IO devices.
 *
 * Sealevel and SeaMAX are registered trademarks of Sealevel Systems
 * Incorporated.
 *
 *  2008-2023 Sealevel Systems, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <termios.h>

#include <seamaxlin.h>
#include <cethernet.h>

#include "TestIt.h"

// Connection configuration settings.
int 		quit = 0;
int             isEth = 0;        //if the connection is Ethernet
int             isLite = 0;       //if the connection is SeaDAC Lite (USB Direct)
char 		name[MAXLENGTH];  //(sealevel_tcp://x.x.x.x)
baud_rates_t 	baud = BRNONE;    //Baud rate of the connection (RTU)
parity_t	parity = P_NONE;  //Parity of the connection (RTU)

CSeaMaxLin 	myModule;         //The actual connection class

int 		myDevice = 0;     //The slave ID
int		model = 0;        //The model number
unsigned char   type = 0;         //Part of model #


// ****************************************************************************
// These are the actual CSeaMax library function calls.  These may... or may
// not be helpful as examples.
// ****************************************************************************


// ----------------------------------------------------------------------------
// Print out connected device info.  Model, BR, etc.
// ----------------------------------------------------------------------------
void devInfo()
{
	fprintf(stderr, "--------------------------------------------\n");
	
	// Information about the currently opened device.  BR and PAR are the
	// same as the connection module, so I'm not going to reprint that.
	if(myDevice > 0)
	{
		fprintf(stderr, "Device Slave ID: %d\t", myDevice);
		fprintf(stderr, "Model Number: %d", model);
		if (model > 400 && model <= 1000)
		{
			printType(type);
		}
		fprintf(stderr, "\n\n");
	}
	// No device is opened.
	else
	{
		fprintf(stderr, "No device open.  You must now open a device.\n\n");
	}
}

// ----------------------------------------------------------------------------
// Print the baud rate in English rather than as an integer.
// ----------------------------------------------------------------------------
void printBaud(baud_rates_t baud)
{
	fprintf(stderr, "Baud Rate: ");

	// All of the baud rates.
	switch(baud)
	{
		case BR1200:    fprintf(stderr, "1200"); break;
		case BR2400:    fprintf(stderr, "2400"); break;
		case BR4800:    fprintf(stderr, "4800"); break;
		case BR9600:    fprintf(stderr, "9600"); break;
		case BR19200:   fprintf(stderr, "19200"); break;
		case BR38400:   fprintf(stderr, "38400"); break;
		case BR57600:   fprintf(stderr, "57600"); break;
		case BR115200:  fprintf(stderr, "115200"); break;
		default:	fprintf(stderr, "?unknown?"); break;
	}

	fprintf(stderr, "\t");
}

// ----------------------------------------------------------------------------
// Print the parity in English rather than as an integer.
// ----------------------------------------------------------------------------
void printParity(parity_t parity)
{
	fprintf(stderr, "Parity: ");

	// Only three possible parities.
	switch(parity)
	{
		case P_NONE:    fprintf(stderr, "NONE"); break;
		case P_ODD:     fprintf(stderr, "ODD"); break;
		case P_EVEN:    fprintf(stderr, "EVEN"); break;
		default:	fprintf(stderr, "?unknown?"); break;
	}

	fprintf(stderr, "\n");
}

// ----------------------------------------------------------------------------
// Print the bridge-type for a particular device.  (M=485, E=Eth, U=USB, etc)
// ----------------------------------------------------------------------------
void printType(unsigned char type)
{
	// Possible bridge types
	switch(type)
	{
		case 0:   fprintf(stderr, "M"); break;  //485
		case 1:   fprintf(stderr, "E"); break;  //Eth
		case 2:   fprintf(stderr, "U"); break;  //Usb
		case 3:   fprintf(stderr, "S"); break;  //232
		case 4:   fprintf(stderr, "N"); break;  //expansion
		default:  break;
	}
}

// ----------------------------------------------------------------------------
// This is the function that attempts to find the list of Ethernet
// devices by using the CEthernet library.  The default is only 5 modules.
// ----------------------------------------------------------------------------
void findEth()
{
	int 		MAXIMUM_DEVICES = 5;  //Only want 5 devices
        CCEthernet      cEthernet;      
        ceth_device_p   arrayOfFoundDevices;
        ceth_device_p   devicePointer;
        int             deviceCount;

        // Allocate a buffer to hold the devices to be found
        arrayOfFoundDevices = cEthernet.Alloc(MAXIMUM_DEVICES);
        
        // Search the network by sending a UDP broadcast to find all
        // Sealevel type ethernet devices
        deviceCount = cEthernet.find_devices(Sealevel_All_Devices, 
                                        MAXIMUM_DEVICES, arrayOfFoundDevices);

        // If we have found at least one device, print
        if(deviceCount > 0)
        {
                fprintf(stderr, "CEthernet Found %d Ethernet Device", deviceCount);
		if (deviceCount > 1) fprintf(stderr, "s");
		fprintf(stderr, ":\n");

                // Set a temporary pointer to the first device in the list
                devicePointer = arrayOfFoundDevices;

                for(int i = 0; i < deviceCount; i++)
                {
                        // Print the device's information and increment the 
                        // pointer to the next device
        		// This points to the first element of the ip byte array.
        		unsigned char *IP = devicePointer->ip_address.c;

        		fprintf(stderr, "  (%s) ", devicePointer->name);
        		fprintf(stderr, "%d.%d.%d.%d\n", IP[0], IP[1], IP[2], IP[3] );
                        devicePointer = devicePointer->next;
                }
        }
	fprintf(stderr, "\n");

	// Free the used memory!
	cEthernet.Free(arrayOfFoundDevices);
}

// ----------------------------------------------------------------------------
// Attempt to make a connection to a module.
// ----------------------------------------------------------------------------
void doOpenMod(char *unused, char *moduleName)
{
	int err = 0;
	char tcp[16] = "sealevel_tcp://";
	char lite[16] = "sealevel_d2x://";

	// Make sure there aren't any already opened.
	doCloseMod(NULL, NULL);

	// This will print an error if device name is unknown or some error occurs.
	err = myModule.Open(moduleName);
	if (err < 0)
	{
		printError(err, NULL);
		return;
	}

	// Update connection info
	memcpy(name, moduleName, MAXLENGTH);

	// Ethernet modules have very odd and limited baudrate configuration
	// capability and I'm not going to even mess with it.
	if (strncmp(moduleName, tcp, 15) == 0)
		isEth = 1;
	// SeaDAC Lite devices use slightly different API calls to access IO
	// also they have no equivalent of serial connection parameters 
	else if (strncmp(moduleName, lite, 15) == 0)
	{
		isLite = 1;
		myDevice = 247;	// Slave ID is not used on SeaDAC Lite devices
		model = strtol(&moduleName[15], NULL, 10);
		type = 2;	// USB
	}
	else
	{
		baud = BR9600;
		parity = P_NONE;
	}

	fprintf(stderr, "Successfully opened module: %s\n\n", name);
}

// ----------------------------------------------------------------------------
// Open a particular device on an already connected module.
// ----------------------------------------------------------------------------
void doOpenDev(char *unused, char *slaveID)
{
	int err = 0;
	int slave = atoi(slaveID);
	seaio_ioctl_s ioctlstruct;

	// Must have a module through which to communicate
	if (name[0] == 0x00)
	{
		fprintf(stderr, "You must first open a module.\n\n");
		return;
	}
	// Make sure slaveID is valid
	if (slave < 1 || slave > 247)
	{
		fprintf(stderr, "Invalid SlaveID: %s\n\n", slaveID);
		return;
	}

	// IOCTL call to determine general device settings.
	err = myModule.Ioctl(slave, IOCTL_READ_COMM_PARAM, &ioctlstruct); 
	if (err < 0)
	{
		printError(err, NULL);
		return;
	}

	// Update internal data and inform user.
	myDevice = slave;
	model = ioctlstruct.u.params.model;
	type = ioctlstruct.u.params.bridge_type;

	// Newer devices (520++ and SeaDAC) use a different modbus command for model
	if (model == 256)
	{
		err = myModule.Ioctl(slave, IOCTL_GET_EXT_CONFIG, &ioctlstruct);
		if (err < 0)
		{
			printError(err, NULL);
			return;
		}
		model = ioctlstruct.u.config.model;
	}

	fprintf(stderr, "Successfully opened %d", model);
	if (model > 400 && model <= 1000)
	{
		printType(type);
	}
	fprintf(stderr, " (slaveID %d) at\n", myDevice);
	printBaud(ioctlstruct.u.params.baud_rate);
	printParity(ioctlstruct.u.params.parity);
	fprintf(stderr, "\n");	
}

// ----------------------------------------------------------------------------
// Close the currently connected module.
// ----------------------------------------------------------------------------
void doCloseMod(char *unused, char *unused2)
{
	// Destroy current module connection.
	myModule.Close();

	// Clear internal data.
	memset(name, 0x00, MAXLENGTH);
	isEth = 0;
	isLite = 0;
	baud = BRNONE;
	parity = P_NONE;
	myDevice = 0;
	model = 0;
	type = 0;
}

// ----------------------------------------------------------------------------
// This function is used to print out module connection information.
// If no modules are open, it will attempt to use the CEthernet library
// to display a list of ethernet modules found through a UDP broadcast.
// ----------------------------------------------------------------------------
void doGetInfo(char *unused, char *unused2)
{
	// If not null, then we have a module open
	if(name[0] != 0x00)
	{
		fprintf(stderr, "Module Name: %s\n", name);
		if (!isEth && !isLite)
		{
			printBaud(baud);
			printParity(parity);
		}
		devInfo();
	}
	// Nothing is open, so offer some suggestions.
	else
	{
		fprintf(stderr, "No module currently connected.");
		fprintf(stderr, "  Please Open a module.\n");
		findEth();
	}
}

// ----------------------------------------------------------------------------
// Set the comm parameters for the module connection.  This is the line from 
// the terminal to the modbus internal bus.  This connection rate/parity must
// be the same as that of the device you wish to use.  So, if you absolutely
// must change your rate/parity, you will need to first change the parameters
// of the device then change the module connection.
// ----------------------------------------------------------------------------
void doSetMod(char *BR, char *PAR)
{
	baud_rates_t  newbaud = (baud_rates_t) atoi(BR);
	parity_t      newparity = (parity_t) atoi(PAR);
	HANDLE        comHandle = -1;

	// Must have a module
	if (name[0] == 0x00)
	{
		fprintf(stderr, "You must first open a module.\n\n");
		return;
	}
	if (isLite)
	{
		fprintf(stderr, "Not supported on SeaDAC Lite modules.\n\n");
		return;
	}
	if (isEth)
	{
		// This technically isn't true, if you ABSOLUTELY must you can
		// broadcast the change through the SET_COMM_PARAM ioctl with
		// a slaveID of 0.  This will also change all the devices
		// chained to that module.
		fprintf(stderr, "Not supported on Ethernet modules.\n\n");
		return;
	}
	// Validate parameters.
	if (newbaud < BR1200 || newbaud > BR115200)
	{
		fprintf(stderr, "Invalid baud rate: %s\n\n", BR);
		return;
	}
	if (newparity < P_NONE || newparity > P_EVEN)
	{
		fprintf(stderr, "Invalid parity: %s\n\n", PAR);
		return;
	}

	comHandle = myModule.getCommHandle();

	if (comHandle > 0)  //validate
	{
		struct termios newtio;            //serial config struct
		bzero(&newtio, sizeof(newtio));   //clear the new config

		// Baud rate selection
		switch (newbaud)
		{
			case BR1200:    newtio.c_cflag = B1200; break;
			case BR2400:    newtio.c_cflag = B2400; break;
			case BR4800:    newtio.c_cflag = B4800; break;
			case BR9600:    newtio.c_cflag = B9600; break;
			case BR19200:   newtio.c_cflag = B19200; break;
			case BR38400:   newtio.c_cflag = B38400; break;
			case BR57600:   newtio.c_cflag = B57600; break;
			case BR115200:  newtio.c_cflag = B115200; break;
			default:	fprintf(stderr, "CRITICAL ERROR! BR Unknown!\n\n"); return;
		}

		//local use and readable
		newtio.c_cflag |= (CLOCAL | CREAD);

		//Parity selection
		if (newparity == P_NONE)
		{
			newtio.c_cflag |= CS8;
			newtio.c_iflag = IGNPAR;
		}
		else
		{
			newtio.c_cflag |= (CS8 | PARENB);
			if (newparity == P_ODD) newtio.c_cflag |= PARODD;
			newtio.c_iflag = INPCK;
		}

		newtio.c_cc[VTIME] = 10;  //1 second timeout on reads.

		tcflush(comHandle, TCIFLUSH);  //clear the line
		tcsetattr(comHandle, TCSANOW, &newtio);  //set this config
	}

	// Now update internal data.
	baud = newbaud;
	parity = newparity;
	myDevice = 0;
	model = 0;
	type = 0;

	// It would be nice to let the user know whats up.
	fprintf(stderr, "Successfully changed connection parameters.\n\n");
}

// ----------------------------------------------------------------------------
// Set device communication parameters.  This is only for one device in
// particular on a chain.  So if you want all your devices to use a specific
// connection rate/parity, you need to configure each device manually.  Also
// there is not support for chaning these settings on Ethernet modules.
// ----------------------------------------------------------------------------
void doSetDev(char *BR, char *PAR)
{
	int err = 0;
	baud_rates_t  newbaud = (baud_rates_t) atoi(BR);
	parity_t      newparity = (parity_t) atoi(PAR);
	seaio_ioctl_s ioctlstruct;

	// Nonsense on a SeaDAC Lite
	if (isLite)
	{
		fprintf(stderr, "Not supported on SeaDAC Lite modules.\n\n");
		return;
	}
	// Must have a module
	if (name[0] == 0x00)
	{
		fprintf(stderr, "You must first open a module.\n\n");
		return;
	}
	// Must have a device
	if (myDevice < 1 || myDevice > 247)
	{
		fprintf(stderr, "You must first open a device.\n\n");
		return;
	}
	if (isEth)
	{
		// This technically isn't true, if you ABSOLUTELY must you can
		// broadcast the change through the SET_COMM_PARAM ioctl with
		// a slaveID of 0.  This will also change all the devices
		// chained to that module.
		fprintf(stderr, "Not supported on Ethernet modules.\n\n");
		return;
	}
	// Validate parameters.
	if (newbaud < BR1200 || newbaud > BR115200)
	{
		fprintf(stderr, "Invalid baud rate: %s\n\n", BR);
		return;
	}
	if (newparity < P_NONE || newparity > P_EVEN)
	{
		fprintf(stderr, "Invalid parity: %s\n\n", PAR);
		return;
	}
	
	// Make the Ioctl call -- first we have to read the current settings
	// so we can get a valid Magic Cookie
	err = myModule.Ioctl(myDevice, IOCTL_READ_COMM_PARAM, &ioctlstruct);
	if (err < 0)
	{
		printError(err, NULL);
		return;
	}

	// Update the settings
	ioctlstruct.u.comms.new_baud_rate = newbaud;
	ioctlstruct.u.comms.new_parity = newparity;
	
	err = myModule.Ioctl(myDevice, IOCTL_SET_COMM_PARAM, &ioctlstruct);
	if (err < 0)
	{
		printError(err, NULL);
		return;
	}

	// Now if we got this far, the device is talking gibberish to this module.
	myDevice = 0;
	model = 0;
	type = 0;
	
	// It would be nice to let the user know whats up.
	fprintf(stderr, "Successfully changed device communication parameters.\n\n");
}

// ----------------------------------------------------------------------------
// Set the slave ID for a particular module.  This can be for 1 to 247.  The
// factory default is 247.
// ----------------------------------------------------------------------------
void doSetDevID(char *unused, char *slaveID)
{
	int err = 0;
	slave_address_t newID = (slave_address_t) atoi(slaveID);
	seaio_ioctl_s ioctlstruct;

	// Nonsense on a SeaDAC Lite
	if (isLite)
	{
		fprintf(stderr, "Not supported on SeaDAC Lite modules.\n\n");
		return;
	}
	// Must have a module
	if (name[0] == 0x00)
	{
		fprintf(stderr, "You must first open a module.\n\n");
		return;
	}
	// Must have a device
	if (myDevice < 1 || myDevice > 247)
	{
		fprintf(stderr, "You must first open a device.\n\n");
		return;
	}
	// Validate parameters.
	if (newID < 1 || newID > 247)
	{
		fprintf(stderr, "Invalid slaveID: %s\n\n", slaveID);
		return;
	}
	
	// Make the Ioctl call -- first we have to read the current settings
	// so we can get a valid Magic Cookie
	err = myModule.Ioctl(myDevice, IOCTL_READ_COMM_PARAM, &ioctlstruct);
	if (err < 0)
	{
		printError(err, NULL);
		return;
	}

	// Update the settings
	ioctlstruct.u.address.new_address = newID;
	
	err = myModule.Ioctl(myDevice, IOCTL_SET_ADDRESS, &ioctlstruct);
	if (err < 0)
	{
		printError(err, NULL);
		return;
	}

	// Now if we got this far, the device has the new slaveID.
	myDevice = newID;
	
	// It would be nice to let the user know whats up.
	fprintf(stderr, "Successfully changed slaveID to %s\n\n", slaveID);
}

// ----------------------------------------------------------------------------
// Read all of the IO on a device.
// ----------------------------------------------------------------------------
void doReadAll(char *unused1, char *unused2)
{
	// Data buffer for IO.  Max size is based on PIO having 96bits I/O.
	unsigned char buffer[12];
	int i, err = 0;
	int in, out, pio;
	in = out = pio = 0;

	// Must have a module
	if (name[0] == 0x00)
	{
		fprintf(stderr, "You must first open a module.\n\n");
		return;
	}
	// Must have a device
	if (myDevice < 1 || myDevice > 247)
	{
		fprintf(stderr, "You must first open a device.\n\n");
		return;
	}

	// Determine how many IO bits there are.
	switch (model)
	{
		case 8111:
		case 8112:
		case 12009:
		case 110: 
		case 120: in = out = 4;     break;
		case 8113:
		case 130: in = 4;           break;
		case 8114:
		case 8115:
		case 9194:
		case 140:
		case 150: out = 4;          break;
		case 160: pio = 32;         break;
		case 170: in = out = 2;     break;
		case 8221:
		case 410:
		case 530: in = out = 16;    break;
		case 8222:
		case 420: in = 16; out = 8; break;
		case 8223:
		case 430:
		case 580: in = 32;          break;
		case 8224:
		case 440: 
		case 540: out = 32;         break;
		case 8225:
		case 450: out = 16;         break;
		case 8126: pio = 32;        break;
		case 462:
		case 463: pio = 96;         break;
		case 8227:
		case 470:
		case 570: in = out = 8;     break;
		case 8232:
		case 520: in = out = 8;     break;
		default:  fprintf(stderr, "Read not supported.\n\n"); return;
	}

	// If applicable, read all Inputs.
	if (in > 0)
	{	
		if (!isLite) err = myModule.Read(myDevice, D_INPUTS, 1, in, buffer);
		else err = myModule.Read(buffer, 1);
		if (err < 0)
		{
			printError(err, NULL);
			return;
		}

		// Now print the data returned byte by byte.
		fprintf(stderr, "INPUTS: (0= inactive, 1= active)\n");
		if (isLite)
			fprintf(stderr, "bank1-0x%02X\t", buffer[0] & 0x0F);
		else if (in < 8)
			fprintf(stderr, "bank1-0x%02X\t", buffer[0]);
		for (i = 0; i < (in/8); i++)
			fprintf(stderr, "bank%d-0x%02X\t", i+1, buffer[i]);
		fprintf(stderr, "\n\n");
	}
	
	// If applicable, read all Outputs.
	if (out > 0)
	{	
		if (!isLite) err = myModule.Read(myDevice, COILS, 1, out, buffer);
		else err = myModule.Read(buffer, 1);
		if (err < 0)
		{
			printError(err, NULL);
			return;
		}

		// Now print the data returned byte by byte.
		fprintf(stderr, "OUTPUTS: (0= inactive, 1= active)\n");
		if (isLite)
			fprintf(stderr, "bank1-0x%02X\t", (buffer[0] >> 4) & 0x0F);
		else if (out < 8)
			fprintf(stderr, "bank1-0x%02X\t", buffer[0]);
		for (i = 0; i < (out/8); i++)
			fprintf(stderr, "bank%d-0x%02X\t", i+1, buffer[i]);
		fprintf(stderr, "\n\n");
	}
	
	// If applicable, read PIO.
	if (pio > 0)
	{	
		if (isLite) err = myModule.GetPIO(buffer);
		else err = myModule.Read(myDevice, SEAMAXPIO, 1, pio, buffer);
		if (err < 0)
		{
			printError(err, NULL);
			return;
		}

		// Now print the data returned byte by byte.
		fprintf(stderr, "PIO:\n");
		for (i = 0; i < (pio/8); i++)
		{
			if (i % 6 == 0) fprintf(stderr, "\n");
			fprintf(stderr, "bank%d-0x%02X\t", i+1, buffer[i]);
		}
		fprintf(stderr, "\n\n");
	}
}

// ----------------------------------------------------------------------------
// Write a byte to a single bank.  On I/O modules, banks are Outputs starting
// with bank 1, and each bank is 8 bits.  Writing a 0 corresponds to an "OFF"
// state for relays and 1 an "ON".  On PIO modules, there is no difference in
// banks configured for Input or Output.  So if you want to write to the first
// bank, you will only see the result if it is configured for output and not
// input.  On PIO modules output logic depends on the device.
// ----------------------------------------------------------------------------
void doWriteBank(char *port, char *datastr)
{
	// Convert input strings to actual values.
	int bank = atoi(port);
	unsigned char data = (unsigned char) atoi(datastr);

	// Data buffer for Output.  In a PIO write, we have to read all first.
	// Then turn around and write back everything, with the modification.
	unsigned char buffer[12];
	int err = 0;
	int out, pio;
	out = pio = 0;

	// Must have a module
	if (name[0] == 0x00)
	{
		fprintf(stderr, "You must first open a module.\n\n");
		return;
	}
	// Must have a device
	if (myDevice < 1 || myDevice > 247)
	{
		fprintf(stderr, "You must first open a device.\n\n");
		return;
	}

	// Determine how many banks there are. (number of bits)
	switch (model)
	{
		case 8111:
		case 8112:
		case 12009:
		case 8114:
		case 8115:
		case 9194:
		case 110:
		case 120: out = 4; break;
		case 130: out = 0; break;
		case 140:
		case 150: out = 4; break;
		case 8126:
		case 160: pio = 32; break;
		case 170: out = 2; break;
		case 8221:
		case 410: out = 16; break;
		case 8222:
		case 420: out = 8;  break;
		case 8223:
		case 430: out = 0;  break;
		case 8224:
		case 440: out = 32; break;
		case 8225:
		case 450: out = 16; break;
		case 462:
		case 463: pio = 96; break;
		case 8227:
		case 470: out = 8;  break;
		case 8232:
		case 520: out = 8;  break;
		default:  fprintf(stderr, "Write not supported.\n\n"); return;
	}

	// Validate bank #
	if (  (bank <= 0) || ( (((bank - 1) * 8) > out) && ((bank * 8) > pio) )  )
	{
		fprintf(stderr, "ERROR: INVALID BANK#-%s\n\n", port);
		return;
	}

	// If applicable, write to Outputs.  (COILS)
	if (out > 0)
	{
		// Starting bit, MODBUS is base 1
		slave_address_t addr = (  ( (bank - 1) * 8 ) + 1  );
		
		// Put the data we want to write in the buffer
		buffer[0] = data;

		// Attempt to write the data and print an error if we get one.
		if (isLite) { buffer[0] = buffer[0] << 4; err = myModule.Write(buffer, 1); }
		else if (out < 8) err = myModule.Write(myDevice, COILS, addr, out, buffer);
		else err = myModule.Write(myDevice, COILS, addr, 8, buffer);
		if (err < 0)
		{
			printError(err, NULL);
			return;
		}
	}

	// If applicable, write to PIO. (First read it!)
	if (pio > 0)
	{
		// Read the current IO of the device.
		if (isLite) err = myModule.GetPIO(buffer);
		else err = myModule.Read(myDevice, SEAMAXPIO, 0, 0, buffer);
		if (err < 0)
		{
			printError(err, NULL);
			return;
		}

		// Modify the desired byte.  Also make a slight pause.
		buffer[bank-1] = data;
		usleep(1000*50);       //50ms (this is necessary for eth type)
	
		// Finally, attempt to write the modified data back.
		if (isLite) err = myModule.SetPIO(buffer);
		else err = myModule.Write(myDevice, SEAMAXPIO, 0, 0, buffer);
		if (err < 0)
		{
			printError(err, NULL);
			return;
		}
	}

	// Everything was a success, thanks for shopping here, come again.
	fprintf(stderr, "Successfully wrote %s to bank%s.\n\n", datastr, port);
}

// ----------------------------------------------------------------------------
// Configure programable options on devices.  PIO and A/D for now.
// ----------------------------------------------------------------------------
void doSetup(char *cw1, char *cw2)
{
	int err = 0;
	unsigned char control1 = (unsigned char) atoi(cw1);
	unsigned char control2 = (unsigned char) atoi(cw2);
	unsigned char buffer[4];
	
	// Must have a module
	if (name[0] == 0x00)
	{
		fprintf(stderr, "You must first open a module.\n\n");
		return;
	}
	// Must have a device
	if (myDevice < 1 || myDevice > 247)
	{
		fprintf(stderr, "You must first open a device.\n\n");
		return;
	}

	// SeaDAC Lite PIO configuration
	if (model == 8126)
	{
		// Translate control word 1 lsbs into high == output, low == input for each bank
		buffer[0] = (control1 & 0x01) ? 0x00 : 0xFF;
		buffer[1] = (control1 & 0x02) ? 0x00 : 0xFF;
		buffer[2] = (control1 & 0x04) ? 0x00 : 0xFF;
		buffer[3] = (control1 & 0x08) ? 0x00 : 0xFF;

		err = myModule.SetPIODirection(buffer);
		if (err < 0)
		{
			printError(err, NULL);
			return;
		}

		fprintf(stderr, "Successfully setup PIO!\n\n");
		return;
	}

	// PIO configuration (eIO)
	if (model == 160)
	{
		// Translate control word 1 lsbs into high == output, low == input for each bank
		buffer[0] = 0;
		buffer[1] = (~control1) & 0x0F;

		err = myModule.Write(myDevice, HOLDINGREG, 3, 1, buffer);
		if (err < 0)
		{
			printError(err, NULL);
			return;
		}

		fprintf(stderr, "Successfully setup PIO!\n\n");
		return;
	}

	// PIO configuration
	if (model == 462 || model == 463)
	{
		seaio_ioctl_s ioctl;

		// Input user data into Ioctl struct
		ioctl.u.pio.config_state.PIO96.channel1 = control1;
		ioctl.u.pio.config_state.PIO96.channel2 = control2;

		// Call the Ioctl with the user's data.
		err = myModule.Ioctl(myDevice, IOCTL_SET_PIO, &ioctl);
		if (err < 0)
		{
			printError(err, NULL);
			return;
		}

		fprintf(stderr, "Successfully setup PIO!\n\n");
		return;
	}

	// A/D configuration
	if (model == 470 || model == 8227)
	{
		adda_config adda;

		// First let's get the current A/D config, so we don't 
		// change the channel ranges.
		err = myModule.Ioctl(myDevice, IOCTL_GET_ADDA_CONFIG, &adda);
		if (err < 0)
		{
			printError(err, NULL);
			return;
		}

		// Set user provided data.
		adda.device.reference_offset = control1;
		adda.device.channel_mode = control2;

		err = myModule.Ioctl(myDevice, IOCTL_SET_ADDA_CONFIG, &adda);
		if (err < 0)
		{
			printError(err, NULL);
			return;
		}

		fprintf(stderr, "Successfully setup A/D!\n\n");
		return;
	}

	fprintf(stderr, "Device configuration not supported.\n\n");
}

// ----------------------------------------------------------------------------
// Set the range on an A/D channel.
// ----------------------------------------------------------------------------
void doSetRange(char *channelstr, char *rangestr)
{
	int err = 0;
	int channels = 0;
	unsigned char channel = (unsigned char) atoi(channelstr);
	unsigned char range   = (unsigned char) atoi(rangestr);
	adda_config   adda;
	unsigned char *ptrconfig;
	
	// Nonsense on a SeaDAC Lite
	if (isLite)
	{
		fprintf(stderr, "Not supported on SeaDAC Lite modules.\n\n");
		return;
	}
	// Must have a module
	if (name[0] == 0x00)
	{
		fprintf(stderr, "You must first open a module.\n\n");
		return;
	}
	// Must have a device
	if (myDevice < 1 || myDevice > 247)
	{
		fprintf(stderr, "You must first open a device.\n\n");
		return;

	}

	switch (model)
	{
		case 8227:
		case 470: channels = 16; break;
		default:  fprintf(stderr, "No A/D channels!\n\n"); return;
	}

	// Valid channel?
	if (channel <= 0 || channel > channels)
	{
		fprintf(stderr, "Channel out of range!\n\n");
		return;
	}

	// Read current config.  So we don't change anything else.
	err = myModule.Ioctl(myDevice, IOCTL_GET_ADDA_CONFIG, &adda);
	if (err < 0)
	{
		printError(err, NULL);
		return;
	}

	ptrconfig = &adda.channels.ch_1;
	ptrconfig[channel - 1] = range;

	// Set the new config.
	err = myModule.Ioctl(myDevice, IOCTL_SET_ADDA_CONFIG, &adda);
	if (err < 0)
	{
		printError(err, NULL);
		return;
	}

	fprintf(stderr, "Successfully set A/D channel%d range.\n\n", channel);
}

// ----------------------------------------------------------------------------
// Read all the analog channels from an A/D capable device.
// ----------------------------------------------------------------------------
void doMonitorAtoD(char *unused, char *unused2)
{
	int err = 0, numChannels = 0;
	unsigned char data[2];
	unsigned int value;
	double percentage;
	char unit[10];
	adda_ext_config extended;
	adda_config config;
	unsigned char *ptrchannels;

	// Nonsense on a SeaDAC Lite
	if (isLite)
	{
		fprintf(stderr, "Not supported on SeaDAC Lite modules.\n\n");
		return;
	}
	// Must have a module
	if (name[0] == 0x00)
	{
		fprintf(stderr, "You must first open a module.\n\n");
		return;
	}
	// Must have a device
	if (myDevice < 1 || myDevice > 247)
	{
		fprintf(stderr, "You must first open a device.\n\n");
		return;
	}

	// Determine if the device even has A/D
	switch(model)
	{
		case 170:
		case 570: numChannels = 8; break;
		case 8227:
		case 470: numChannels = 16; break;
		default: fprintf(stderr, "Device has no A/D channels.\n\n");
			 break;
	}

	// Read configuration information
	err = myModule.Ioctl(myDevice, IOCTL_GET_ADDA_CONFIG, &config);
	if (err < 0)
	{
		printError(err, NULL);
		return;
	}

	// Easy access to channel config data.
	ptrchannels = &config.channels.ch_1;

	// Extra configuration information.
	err = myModule.Ioctl(myDevice, IOCTL_GET_ADDA_EXT_CONFIG, &extended);
	if (err < 0)
	{
		printError(err, NULL);
		return;
	}

	// Other configurations.
	if (config.device.channel_mode == DIFFERENTIAL || 
			config.device.channel_mode == CURRENT_LOOP)
		numChannels = 8;

	// Read A/D channel values one at a time.
	for (int i = 0; i < numChannels; i++)
	{
		memset(data, 0, sizeof(data));
	
		err = myModule.Read(myDevice, INPUTREG, i+1, 1, data);
		if (err < 0)
		{
			printError(err, NULL);
			return;
		}

		// Convert from the modbus 2 bytes to an actual value
		value = (data[0] << 8) + (data[1]);
		
		if (ptrchannels[i] == ZERO_TO_FIVE ||
					ptrchannels[i] == ZERO_TO_TEN)
		{
			percentage = value / 4095.0;
		}
		else if (ptrchannels[i] == PLS_MIN_FIVE ||
					ptrchannels[i] == PLS_MIN_TEN)
		{
			// Adjust for the sign bit
			if (value & 0x800)
			{
				percentage = ((double)(value & 0x7FF)-2047.0) 
						/ 2047.0;
			}
                
			// If the sign bit was not set, then the value is 
			// positive and the extra manipulation isn't necessary
			else
			{
				percentage = value / 2047.0;
			}

		}
		else
		{
			fprintf(stderr, "ERROR: Unknown A/D config\n\n");
			return;
		}

		// If the hardware 10x gain jumper is enabled, the the maximum 
		// values range from (0,0.5) or (0, 1)
		if (extended.ad_multiplier_enabled)
	        {
			percentage = percentage / 10;
		}

		// In current loop mode, the two differential inputs are 
		// connected through a 249 Ohm precision resistor.  Since 
		// voltage = current * resistance, the current is calculated 
		// as the voltage divided by the resistance.
		if (config.device.channel_mode == CURRENT_LOOP)
		{
			percentage = percentage / 249.0 * 1000.0;
			strcpy(unit, "mAmps");
		}
		else
		{
			strcpy(unit, "Volts");
		}

		fprintf(stderr, "A/D Channel%d: ", i+1);
		fprintf(stderr, "(%02X%02X) ", data[0], data[1]);
		switch(ptrchannels[i])
		{
			case ZERO_TO_FIVE:
			case PLS_MIN_FIVE: 
			  fprintf(stderr, "%.2f %s\n", percentage * 5.0, unit);
			  break;
			case ZERO_TO_TEN:
			case PLS_MIN_TEN:
			  fprintf(stderr, "%.2f %s\n", percentage * 10.0, unit);
			  break;
		}
	}

	fprintf(stderr, "\n");
}

// ----------------------------------------------------------------------------
// Write a value to a D/A channel.  The value is from 0x000 to 0xFFF.
// ----------------------------------------------------------------------------
void doWriteDtoA(char *which, char *what)
{
	int err = 0, numChannels = 0;
	int channel = atoi(which);
	unsigned short swap = (unsigned short) atoi(what);
	unsigned char data[2];

	data[0] = ((swap >> 8) & 0x00FF);
	data[1] = ((swap >> 0) & 0x00FF);

	// Nonsense on a SeaDAC Lite
	if (isLite)
	{
		fprintf(stderr, "Not supported on SeaDAC Lite modules.\n\n");
		return;
	}
	// Must have a module
	if (name[0] == 0x00)
	{
		fprintf(stderr, "You must first open a module.\n\n");
		return;
	}
	// Must have a device
	if (myDevice < 1 || myDevice > 247)
	{
		fprintf(stderr, "You must first open a device.\n\n");
		return;
	}

	switch(model)
	{
		case 8227:
		case 470: numChannels = 2; break;
		default:  break;
	}

	// Must be valid D/A channel number
	if (channel <= 0 || channel > numChannels)
	{
		fprintf(stderr, "Invalid D/A channel: %d\n\n", channel);
		return;
	}

	// Attempt to write the D/A value
	err = myModule.Write(myDevice, HOLDINGREG, channel, 1, data);
	if (err < 0)
	{
		printError(err, NULL);
		return;
	}

	fprintf(stderr, "Wrote %02X %02X ", data[0], data[1]); 
	fprintf(stderr, "to channel%d.\n\n", channel);
}

// ----------------------------------------------------------------------------
// Knight-Rider style scan.  Turn each bit or relay on and then off and move to
// the next bit or relay.
// ----------------------------------------------------------------------------
void doScan(char *unused, char *maxvalue)
{
	int reps = atoi(maxvalue);
	int i, out = 0;
	seaio_type_t type = COILS;
	unsigned char buffer[12];

	// Must have a module
	if (name[0] == 0x00)
	{
		fprintf(stderr, "You must first open a module.\n\n");
		return;
	}
	// Must have a device
	if (myDevice < 1 || myDevice > 247)
	{
		fprintf(stderr, "You must first open a device.\n\n");
		return;
	}

	switch (model)
	{
		case 8221:
		case 410: 
		case 530: out = 2;  break;
		case 8222:
		case 420: out = 1;  break;
		case 8224:
		case 440: 
		case 540: out = 4;  break;
		case 8225:
		case 450: out = 2;  break;
		case 462:
		case 463: type = SEAMAXPIO; out = 12; break;
		case 160: type = SEAMAXPIO; out = 4; break;
		case 8827:
		case 470: 
		case 570: out = 1;  break;
		case 8232:
		case 520: out = 1;  break;
		default:  fprintf(stderr, "Scan not supported.\n\n"); return;
	}

	// Repeat this pattern
	for (i = 0; i < reps; i++) 
	{
		int j;
		for (j = 0; j < (8 * out); j++)
		{
			memset(buffer, 0x00, sizeof(buffer));	
			buffer[j / 8] = ( 0x01 << (j % 8) );
			myModule.Write(myDevice, type, 1, out*8, buffer);
			usleep(1000*50);  // 50ms pause
		
		}
	
		for (j = (j - 1); j >= 0; j--)
		{
			memset(buffer, 0x00, sizeof(buffer));	
			buffer[j / 8] = ( 0x01 << (j % 8) );
			myModule.Write(myDevice, type, 1, out*8, buffer);
			usleep(1000*50);  // 50ms pause
		}
	}

	// Clean up... turn everything off.
	memset(buffer, 0x00, sizeof(buffer));	
	myModule.Write(myDevice, type, 1, out*8, buffer);

	fprintf(stderr, "Completed %d cycles.\n\n", reps);
}

// ----------------------------------------------------------------------------
// Quit this test application.
// ----------------------------------------------------------------------------
void doQuit(char *unused, char *unused2)
{
	// Closing the module isn't strictly necessary, but eh?
	doCloseMod(NULL, NULL);
	quit = 1;
}

// ----------------------------------------------------------------------------
// Error printing.  First byte of the buffer in modbus comms is the exception
// code, so we may need that too.
// ----------------------------------------------------------------------------
void printError(int error, const char *buff)
{
	switch (error)
	{
		case -EBUSY:
			fprintf(stderr, "Returned error: EBUSY\n\n");
			break;		

		case -EBADF:
			fprintf(stderr, "Returned error: EBADF\n\n"); 
			break;		
			
		case -ENOMEM:
			fprintf(stderr, "Returned error: ENOMEM\n\n");
			break;		

		case -EPERM:
			fprintf(stderr, "Returned error: EPERM\n\n");
			break;		
		
		case -EINVAL:
			fprintf(stderr, "Returned error: EINVAL\n\n");
			break;		
			
		case -ENODEV:
			fprintf(stderr, "Returned error: ENODEV\n\n");
			break;		
			
		case -EFAULT:
			fprintf(stderr, "Returned error: EFAULT ");
			if (buff != NULL) fprintf(stderr, "(MODBUS Exception:%d)", buff[0]);
			fprintf(stderr, "\n\n");
			break;		
			
		case -ENAMETOOLONG:
			fprintf(stderr, "Returned error: ENAMETOOLONG\n\n");
			break;		
			
		default:
			fprintf(stderr, "Returned unknown error: %i\n\n", error);
	}
	return;
}
// ****************************************************************************
// This is recycled code.  Everything following this comment simply handles
// the interface... be that batch operation or event driven.
// ****************************************************************************
int main(int argc, char * argv[])
{
	char buff[3*MAXLENGTH];
	FILE* InputFile = stdin;  //Only need stdin if no input file supplied.

	if (argc > 1)   // input file
	{
		if ((InputFile = fopen(argv[1], "r")) == NULL) // invalid
		{
			printf( "Error:Could not open file %s.\n" , argv[1] );
			InputFile = stdin;
		}
	} 

	if (InputFile == stdin)
		ShowMenu();
	
	while (!quit && fgets(buff, sizeof(buff), InputFile)) 
	{
		if (InputFile == stdin)
			ShowMenu();

		Command *cmd = findCommand(buff);
		
		if (cmd) 
		{
			char arg1[MAXLENGTH], arg2[MAXLENGTH];

			// Simply get the arguments if they weren't already supplied.
			if (cmd->ArgCount == 1 && !getOneArg(buff + 1, arg1)) 
			{
				fprintf(stderr, "Input required: --\n");
				ShowLine(cmd->Arg1);
				fscanf(stdin, "%s", arg1);
				fgetc(stdin);
				fprintf(stderr, "\n");
			} 
			// Same as before.  These should never show up in Batch operation.
			else if (cmd->ArgCount == 2 && 
					!getTwoArgs(buff + 1, arg1, arg2)) 
			{
				fprintf(stderr, "Input required: --\n");
				ShowLine(cmd->Arg1);
				fscanf(stdin, "%s", arg1);
				ShowLine(cmd->Arg2);
				fscanf(stdin, "%s", arg2);
				fgetc(stdin);
				fprintf(stderr, "\n");
			}

			if (cmd->ArgCount == 0)
			{
				if (InputFile != stdin) fprintf(stderr, "%c:\n", cmd->cmd);
				cmd->func(NULL, NULL);
			}
			else if (cmd->ArgCount == 1)
			{
				if (InputFile != stdin) fprintf(stderr, "%c %s:\n", cmd->cmd, arg1);
				cmd->func(NULL, arg1);
			}
			else
			{
				if (InputFile != stdin) fprintf(stderr, "%c %s %s:\n", cmd->cmd, arg1, arg2);
				cmd->func(arg1, arg2);
			}
		}
	}

	// Don't leak memory!
	if (InputFile != stdin) fclose(InputFile);

	return 0;
}

char * Breakline(int LineLength, char * Destination, char * Source)
{	
	// do some stuff to break up the lines
	if (Source == NULL)
		return NULL;
	if ((unsigned int)LineLength > strlen(Source)) {
		strcpy(Destination, Source);
		return NULL;
	}

	char Buffer[MAXLENGTH];
	char * pBuffer = Buffer;
	strcpy(pBuffer, Source);

	while (!isspace(*(pBuffer + LineLength))) LineLength--;
	pBuffer += LineLength;
	
	*pBuffer='\0';

	strcpy(Destination, Buffer);

	return (Source + (LineLength * sizeof(char)) + 1);
}

void ShowMenu(void) 
{
	int line = 0, linelength = LINELENGTH - 4;
	char buff[MAXLENGTH];
	char *placeholder = buff;
	char linebuff[LINELENGTH];
	char *shortline = linebuff;
	system("clear");

	fprintf(stderr, "--------------------------------------------\n");
	fprintf(stderr,"Menu for LinSSD (Command line operation) \n");
	
	for (unsigned int i = 0; i < (sizeof(cmds) / sizeof(Command)); i++) { // run through commands
		if (strlen(cmds[i].Description) < (unsigned int)linelength) // short line, this is easy
			fprintf(stderr,"%c - %s\n",cmds[i].cmd,cmds[i].Description);
		else { // long line
			line = 0;
			placeholder = buff;
			strcpy(placeholder, cmds[i].Description);
			do {
				placeholder = Breakline(linelength, shortline, placeholder);
				fprintf(stderr,"%c %c %s\n",(line == 0?cmds[i].cmd:' '),(line == 0?'-':' '),shortline);
				line++;
			} while (placeholder !=NULL);// end while (going through the long line)
		} // end else (long line)
	} // end for (run through commands)

	fprintf(stderr, "\n");
}


Command *findCommand(const char *buff) 
{
	unsigned int i;
	char cmd;

	// Ignore comment lines and whitespace or NULL commands.
	if (buff[0] == '#' || buff[0] == ' ' || buff[0] == '\n') return NULL;
	
	cmd = toupper(buff[0]);
	for (i = 0; i < (sizeof(cmds) / sizeof(Command)); i++) 
	{
		if (cmds[i].cmd == cmd)
			return cmds + i;
	}

	fprintf(stderr, "Unknown Command '%c'\n", cmd);
	return NULL;
}

int getOneArg(const char *buff, char *result) 
{
	return sscanf(buff, "%s", result) == 1;  //Watch out for spaces in strings...
}

int getTwoArgs(const char *buff, char *result1, char *result2) 
{
	return sscanf(buff, "%s %s", result1, result2) == 2;   //Spaces in strings...
}

void ShowLine(char * instr)
{
	int line = 0, linelength = LINELENGTH;
	char buff[MAXLENGTH];
	char *placeholder = buff;
	char linebuff[LINELENGTH];
	char *shortline = linebuff;

	if (strlen(instr) < (unsigned int)linelength) // short line, this is easy
		fprintf(stderr,"%s--> ", instr);
	else { // long line
		line = 0;
		placeholder = buff;
		strcpy(placeholder, instr);
		do {
			placeholder = Breakline(linelength, shortline, placeholder);
			fprintf(stderr,"%s%c", shortline, placeholder!=NULL?'\n':':');
			line++;
		} while (placeholder !=NULL);// end while
	} // end else (long line)
}

