/*
 * rwio.cpp
 * SeaMAX for Linux Example Code
 *
 * This C++ code demonstrates the reading and writing of SeaDAC Lite 8126
 * programmable inputs and outputs.
 *
 * Sealevel and SeaMAX are registered trademarks of Sealevel Systems
 * Incorporated.
 *
 *  2009-2017 Sealevel Systems, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// Inclusion of the SeaMAX library header is required
#include <seamaxlin.h>

int main(int argc, char * argv[])
{
	int i = 0;
	unsigned char data[4] = {0, 0, 0, 0};

	// Create the SeaMAX object for accessing SeaDAC modules
	// NOTE the difference in the OOP class name (CSeaMaxLin) and the
	// nonOOP struct pointer (SeaMaxLin)
	CSeaMaxLin myModule;

	// Create the module string for use with Open(...)
	// APPLICATION NOTE:
	// "sealevel_d2x://Product Name"
	const char *portString = "sealevel_d2x://8126";

	// Open the SeaDAC module by calling Open() from the library
	int result = myModule.Open(const_cast<char *>(portString));
	if (result != 0)
	{
		printf("ERROR: Open failed, Returned %d\n", result);
		return -1;
	}

	// Configuration data for 8126 TTL is grouped by banks of 8 IO
	// Each byte indicates direction (0 is output, non-zero is input)
	data[0] = 0x00; //OUT
	data[1] = 0x00; //OUT
	data[2] = 0xFF; //IN
	data[3] = 0xFF; //IN
	
	result = myModule.SetPIODirection(data);
	if (result < 0)
	{
		printf("ERROR: Error reading IO Direction, Returned %d\n", result);
		return -1;
	}
	else
	{
		printf("IO Direction(W): %02X%02X%02X%02X\n", data[0], data[1], data[2], data[3]);
	}

	// Clear data structure, to be absolutely sure we configured the device as expected
	for (i = 0; i < 4; i++)
	{
		data[i] = 0x00;
	}

	// Now read back the actual configuration (could do a double check here if we wanted)	
	result = myModule.GetPIODirection(data);
	if (result < 0)
	{
		printf("ERROR: Error reading IO Direction, Returned %d\n", result);
		return -1;
	}
	else
	{
		printf("IO Direction(R): %02X%02X%02X%02X\n", data[0], data[1], data[2], data[3]);
	}

	// Set a pattern for the outputs
	for(i = 0; i < 4; i++)
	{
		data[i] = 0xA5;
	}

	// Configure state of output enabled IO (setting an input has no effect)
	// There is 1 bit for each IO point, each byte corresponds to a bank of IO.
	result = myModule.SetPIO(data);
	if (result < 0)
	{
		printf("ERROR: Error writing IO State, Returned %d\n", result);
		return -1;
	}
	else
	{
		printf("IO State(W): %02X%02X%02X%02X\n", data[0], data[1], data[2], data[3]);
	}

	// Read back the current state of the IO (output state as well as input state)
	// There is 1 bit for each IO point, each byte corresponds to a bank of IO.
	result = myModule.GetPIO(data);
	if (result < 0)
	{
		printf("ERROR: Error reading IO State, Returned %d\n", result);
		return -1;
	}
	else
	{
		printf("IO State(R): %02X%02X%02X%02X\n", data[0], data[1], data[2], data[3]);
	}
		
	// Close the connection and clean up the memory used by SeaMAX
	myModule.Close();
	
	return 0;
}
