/*
 * rwio.cpp
 * SeaMAX for Linux Example Code
 *
 * This C++ code demonstrates the reading and writing of SeaIO 410
 * digital inputs and outputs.
 *
 * Sealevel and SeaMAX are registered trademarks of Sealevel Systems
 * Incorporated.
 *
 *  2006-2017 Sealevel Systems, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>

// Inclusion of the SeaMAX library header is required
#include <seamaxlin.h>

int main(int argc, char * argv[])
{

	// Create the SeaMAX object for accessing SeaIO modules
	// NOTE the difference in the OOP class name (CSeaMaxLin) and the
	// nonOOP struct pointer (SeaMaxLin)
	CSeaMaxLin myModule;

	// A handle to the connection f.o.p.  (Only valid on RTU, not TCP)
	HANDLE comHandle;

	// Two termios structs, one to hold the current setup and one to
	// change so we can talk with our device.
	struct termios original, newtio;

	// Create the module string for use with SeaMAX Open(...)
	//
	// APPLICATION NOTE:
	//     
	//  "sealevel_rtu://dev/DEVNAME" is used for RS232, RS485, and USB 
	//  connections (DEVNAME is the special device file under /dev/)
	//
	//  "sealevel_tcp://xxx.xxx.xxx.xxx:yyy" is used for TCP/IP connections
	//  (where x's represent the ip address and y's the port number)
	//  By default your device will be DHCP enabled and use port 502.
	//  You may choose not to enter the port number and 502 will be used
	//  by default.  You may optionally enter your device name, if you have
	//  configured the device in this way.
	//
	const char *portString = "sealevel_rtu://dev/ttyUSB0";

	// Each SeaIO module has a modbus compliant slave ID.  By default, 
	// modules are shipped with the hardware ID switched to zero, 
	// indicating a slave ID of 247.  Adjusting the hardware switch 
	// to any other address indicates that module will function at the 
	// switch's setting (i.e. a hardware switch setting of 2 indicates
	// that SeaIO module's slave ID is 2).  
	slave_address_t	slaveId = 1;

	// SeaMAX conforms to modbus addressing and function definitions.  
	// (Modbus addressing is one-indexed).  To read, for example, from 
	// inputs 3 to 5, use the following code: type = D_INPUTS, start = 3,
	// range = 3
	seaio_type_t	type;
	address_loc_t	start;
	address_range_t	range;

	// Data is a two byte buffer where the data to be written or read will
	// be stored.  The buffer should be as large as or larger than the
	// number of expected bytes of data for the particular read function.
	// In the case of D_INPUTS, one byte of data is expected for each 8
	// inputs.
	unsigned char data[2] = {0, 0};

	// Open the SeaIO module by calling Open() from the SeaMAX library
	int result = myModule.Open(const_cast<char *>(portString));
	if (result != 0)
	{
		printf("ERROR: Open failed, Returned %d\n", result);
		return -1;
	}

	// Get the host computer communication parameters (only applicable on
	// RTU type connections (RS232, RS485, USB).  ie. Not for TCP type.
	// NOTE:  This is only necessary if you have configured your device to
	// use different communication parameters than the defaults.  Also note
	// that the comm parameters shown below are the defaults, if you have
	// used the IOCTLS to change the defaults, you should use a routine
	// something like this to be able to change your comm port, so you can
	// talk with your device.
	comHandle = myModule.getCommHandle();

	if (comHandle > 0)  //only on RTU type
	{
		tcgetattr(comHandle, &original);  //save the current config
		bzero(&newtio, sizeof(newtio));   //clear the new config
	
		//9600 baud, 8 data bits, local use, readable
		newtio.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
		newtio.c_iflag = IGNPAR;  //ignore parity errors
		newtio.c_oflag = 0;       //raw data
		newtio.c_lflag = 0;       //enable canonical output
		newtio.c_cc[VTIME] = 10;  //1 second timeout on reads.

		tcflush(comHandle, TCIFLUSH);  //clear the line
		tcsetattr(comHandle, TCSANOW, &newtio);  //set this config
	}
	
	// Read the 16 inputs on the module into the data structure
	type	= D_INPUTS;
	start	= 1;
	range	= 16;

	result = myModule.Read(slaveId, type, start, range, data);

	if (result < 0)
	{
		printf("ERROR: Error reading inputs, Returned %d\n", result);
		return -1;
	}
	else
	{
		printf("Read Inputs: %02X%02X\n", data[0], data[1]);
	}

	// Write the value 0x55 out to the outputs, effectively turning them 
	// on or off in an alternating pattern.
	type	= COILS;
	start	= 1;
	range	= 16;

	memset(&data, 0x55, sizeof(data));

	result = myModule.Write(slaveId, type, start, range, data);

	if (result < 0)
	{
		printf("ERROR: Error writing outputs, Returned %d\n", result);
		return -1;
	}
	else
	{
		printf("Write Outputs: %02X%02X\n", data[0], data[1]);
	}

	// Lastly, read in the current state of the outputs.  We should see 
	// the 0x55 pattern that we have recently written out
	memset(&data, 0, sizeof(data));

	result = myModule.Read(slaveId, type, start, range, data);

	if (result < 0)
	{
		printf("ERROR: Error reading outputs, Returned %d\n", result);
		return -1;
	}
	else
	{
		printf("Read Outputs: %02X%02X\n", data[0], data[1]);
	}

	// Close the connection and clean up the memory used by SeaMAX
	myModule.Close();
	
	// Return your comm port to it's original configuration.	
	if (comHandle > 0)  //only on RTU type
	{
		tcflush(comHandle, TCIFLUSH);  //clear the line
		tcsetattr(comHandle, TCSANOW, &original);  //set this config
	}

	return 0;
}
