/*
 * rwpio.cpp
 * SeaMAX for Linux Example Code
 *
 * This C++ code demonstrates the reading and writing of SeaIO 46X programmable 
 * inputs and outputs.  This sample code also makes use of the IOCTL_SET_PIO 
 * command to configure pins as outputs and inputs.
 *
 * Sealevel and SeaMAX are registered trademarks of Sealevel Systems
 * Incorporated.
 *
 *  2006-2017 Sealevel Systems, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>

// Inclusion of the SeaMAX library header is required
#include <seamaxlin.h>

int main(int argc, char * argv[])
{

	// Create the SeaMAX object for accessing SeaIO modules
	// NOTE the difference in the OOP class name (CSeaMaxLin) and the
	// nonOOP struct pointer (SeaMaxLin)
	CSeaMaxLin myModule;

	// A handle to the connection f.o.p.  (Only valid on RTU, not TCP)
	HANDLE comHandle;

	// Two termios structs, one to hold the current setup and one to
	// change so we can talk with our device.
	struct termios original, newtio;

	// Create the module string for use with SeaMAX Open(...)
	//
	// APPLICATION NOTE:
	//     
	//  "sealevel_rtu://dev/DEVNAME" is used for RS232, RS485, and USB 
	//  connections (DEVNAME is the special device file under /dev/)
	//
	//  "sealevel_tcp://xxx.xxx.xxx.xxx:yyy" is used for TCP/IP connections
	//  (where x's represent the ip address and y's the port number)
	//  By default your device will be DHCP enabled and use port 502.
	//  You may choose not to enter the port number and 502 will be used
	//  by default.  You may optionally enter your device name, if you have
	//  configured the device in this way.
	//
	const char *portString = "sealevel_rtu://dev/ttyUSB0";

	// Each SeaIO module has a modbus compliant slave ID.  By default, 
	// modules are shipped with the hardware ID switched to zero, 
	// indicating a slave ID of 247.  Adjusting the hardware switch 
	// to any other address indicates that module will function at the 
	// switch's setting (i.e. a hardware switch setting of 2 indicates
	// that SeaIO module's slave ID is 2).  
	slave_address_t	slaveId = 1;

	// SeaMAX conforms to modbus addressing and function definitions.  
	// (Modbus addressing is one-indexed).  To read, for example, from 
	// inputs 3 to 5, use the following code: type = D_INPUTS, start = 3,
	// range = 3
	seaio_type_t	type;
	address_loc_t	start;
	address_range_t	range;

	// This is the struct used in the majority of IOCTL calls.  For this
	// example code, it is used to configure the port directions.
	seaio_ioctl_s 	ioctl;
	
	// This is the data buffer used in sending and recieving from the 46X.
	// There are 96 bits of IO or 12 bytes.
	unsigned char data[12];

	// Open the SeaIO module by calling Open() from the SeaMAX library
	int result = myModule.Open(const_cast<char *>(portString));
	if (result != 0)
	{
		printf("ERROR: Open failed, Returned %d\n", result);
		return -1;
	}

	// Get the host computer communication parameters (only applicable on
	// RTU type connections (RS232, RS485, USB).  ie. Not for TCP type.
	// NOTE:  This is only necessary if you have configured your device to
	// use different communication parameters than the defaults.  Also note
	// that the comm parameters shown below are the defaults, if you have
	// used the IOCTLS to change the defaults, you should use a routine
	// something like this to be able to change your comm port, so you can
	// talk with your device.
	comHandle = myModule.getCommHandle();

	if (comHandle > 0)  //only on RTU type
	{
		tcgetattr(comHandle, &original);  //save the current config
		bzero(&newtio, sizeof(newtio));   //clear the new config
	
		//9600 baud, 8 data bits, local use, readable
		newtio.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
		newtio.c_iflag = IGNPAR;  //ignore parity errors
		newtio.c_oflag = 0;       //raw data
		newtio.c_lflag = 0;       //enable canonical output
		newtio.c_cc[VTIME] = 10;  //1 second timeout on reads.

		tcflush(comHandle, TCIFLUSH);  //clear the line
		tcsetattr(comHandle, TCSANOW, &newtio);  //set this config
	}

	// For this example it has been decided to configure all ports as
	// outputs.  Channel one configures the first six ports (bytes) and 
	// channel two configures the second six ports (bytes).  Bits 0 through
	// 5 control the direction of ports (bytes) 1 through 6.  A bit value of
	// 1 indicates an input and a bit value of 0 indicates an output.
	ioctl.u.pio.config_state.PIO96.channel1 = 0x00;  //All 6 ports outputs
	ioctl.u.pio.config_state.PIO96.channel2 = 0x3F;  //All 6 ports inputs

	result = myModule.Ioctl(slaveId, IOCTL_SET_PIO, &ioctl);

	if (result < 0)
	{
		printf("ERROR: Configuring PIO, Returned %d\n", result);
		return -1;
	}
	
	// By writing a hex value of 55 to all output ports, the device will
	// effectively have each bit toggled in an alternating pattern.  Since
	// the 46X uses negative logic, a 0 is equivalent to roughly +5V and a
	// 1 to roughly 0V.
	memset(data, 0x55, sizeof(data));

	// A PIO write call expects a data buffer with 12 bytes of data to be
	// written to the PIO device, therefore start and range values are
	// unecessary and may be set to whatever you please.
	type  = SEAMAXPIO;
	start = 0;
	range = 0;

	// A Write() call will only write to the PIO pins set as output.
	// PIO pins set as inputs discard the data to be written and continue
	// to remain inputs.
	result = myModule.Write(slaveId, type, start, range, data);

	if (result < 0)
	{
		printf("ERROR: Writing to PIO, Returned %d\n", result);
		return -1;
	}
	else
	{
		printf("PIO Write: %02X", data[0]);
		printf("%02X",            data[1]);
		printf("%02X",            data[2]);
		printf("%02X",            data[3]);
		printf("%02X",            data[4]);
		printf("%02X",            data[5]);
		printf("%02X",            data[6]);
		printf("%02X",            data[7]);
		printf("%02X",            data[8]);
		printf("%02X",            data[9]);
		printf("%02X",            data[10]);
		printf("%02X\n",          data[11]);
	}

	// Reading is similar to writing.  Simply supply a buffer of at least
	// 12 bytes, and the negative logic values will be returned.  This read
	// should return the 0x55 for all ports configured as outputs.
	memset(data, 0, sizeof(data));
	result = myModule.Read(slaveId, type, start, range, data);

	if (result < 0)
	{
		printf("ERROR: Writing to PIO, Returned %d\n", result);
		return -1;
	}
	else
	{
		printf("PIO Read: %02X",  data[0]);
		printf("%02X",            data[1]);
		printf("%02X",            data[2]);
		printf("%02X",            data[3]);
		printf("%02X",            data[4]);
		printf("%02X",            data[5]);
		printf("%02X",            data[6]);
		printf("%02X",            data[7]);
		printf("%02X",            data[8]);
		printf("%02X",            data[9]);
		printf("%02X",            data[10]);
		printf("%02X\n",          data[11]);
	}

	// Close the connection and clean up the memory used by SeaMAX
	myModule.Close();

	// Return your comm port to it's original configuration.	
	if (comHandle > 0)  //only on RTU type
	{
		tcflush(comHandle, TCIFLUSH);  //clear the line
		tcsetattr(comHandle, TCSANOW, &original);  //set this config
	}

	return 0;
}
