/*
 * setpio.cpp
 * SeaMAX for Linux Example Code
 *
 * This C++ code demonstrates the configuration of SeaIO 46X programmable 
 * inputs and outputs.  NOTE:  The method demonstrated in this example code is 
 * slightly different than that used in the Read/Write PIO example code.  The 
 * RW example code uses an Ioctl call to configure port direction, but this 
 * only holds as long as the device is powered.  As soon as the device loses 
 * power, it will revert to the settings previously written to the holding regs,
 * as this example demonstrates.  Also note that you may configure the power on
 * presets for ports configured as outputs.
 *
 * Sealevel and SeaMAX are registered trademarks of Sealevel Systems
 * Incorporated.
 *
 *  2006-2017 Sealevel Systems, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>

// Inclusion of the SeaMAX library header is required
#include <seamaxlin.h>

void PrintmoduleConfiguration(unsigned char *readHoldRegs);

int main(int argc, char * argv[])
{

	// Create the SeaMAX object for accessing SeaIO modules
	// NOTE the difference in the OOP class name (CSeaMaxLin) and the
	// nonOOP struct pointer (SeaMaxLin)
	CSeaMaxLin myModule;

	// A handle to the connection f.o.p.  (Only valid on RTU, not TCP)
	HANDLE comHandle;

	// Two termios structs, one to hold the current setup and one to
	// change so we can talk with our device.
	struct termios original, newtio;

	// Create the module string for use with SeaMAX Open(...)
	//
	// APPLICATION NOTE:
	//     
	//  "sealevel_rtu://dev/DEVNAME" is used for RS232, RS485, and USB 
	//  connections (DEVNAME is the special device file under /dev/)
	//
	//  "sealevel_tcp://xxx.xxx.xxx.xxx:yyy" is used for TCP/IP connections
	//  (where x's represent the ip address and y's the port number)
	//  By default your device will be DHCP enabled and use port 502.
	//  You may choose not to enter the port number and 502 will be used
	//  by default.  You may optionally enter your device name, if you have
	//  configured the device in this way.
	//
	const char *portString = "sealevel_rtu://dev/ttyUSB0";

	// Each SeaIO module has a modbus compliant slave ID.  By default, 
	// modules are shipped with the hardware ID switched to zero, 
	// indicating a slave ID of 247.  Adjusting the hardware switch 
	// to any other address indicates that module will function at the 
	// switch's setting (i.e. a hardware switch setting of 2 indicates
	// that SeaIO module's slave ID is 2).  
	slave_address_t	slaveId = 1;

	// SeaMAX conforms to modbus addressing and function definitions.  
	// (Modbus addressing is one-indexed).  To read, for example, from 
	// inputs 3 to 5, use the following code: type = D_INPUTS, start = 3,
	// range = 3
	seaio_type_t	type;
	address_loc_t	start;
	address_range_t	range;

	// This data buffer will be used to read and write holding regs.
	// There are 7 holding registers, and each is made up of two bytes.
	unsigned char data[14];
	int i;

	// Open the SeaIO module by calling Open() from the SeaMAX library
	int result = myModule.Open(const_cast<char *>(portString));
	if (result != 0)
	{
		printf("ERROR: Open failed, Returned %d\n", result);
		return -1;
	}

	// Get the host computer communication parameters (only applicable on
	// RTU type connections (RS232, RS485, USB).  ie. Not for TCP type.
	// NOTE:  This is only necessary if you have configured your device to
	// use different communication parameters than the defaults.  Also note
	// that the comm parameters shown below are the defaults, if you have
	// used the IOCTLS to change the defaults, you should use a routine
	// something like this to be able to change your comm port, so you can
	// talk with your device.
	comHandle = myModule.getCommHandle();

	if (comHandle > 0)  //only on RTU type
	{
		tcgetattr(comHandle, &original);  //save the current config
		bzero(&newtio, sizeof(newtio));   //clear the new config
	
		//9600 baud, 8 data bits, local use, readable
		newtio.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
		newtio.c_iflag = IGNPAR;  //ignore parity errors
		newtio.c_oflag = 0;       //raw data
		newtio.c_lflag = 0;       //enable canonical output
		newtio.c_cc[VTIME] = 10;  //1 second timeout on reads.

		tcflush(comHandle, TCIFLUSH);  //clear the line
		tcsetattr(comHandle, TCSANOW, &newtio);  //set this config
	}

	// We wish to access the holding regs, starting with the third.
	// Appendix C in your device manual shows the register layout on
	// 46X devices.
	//
	// There are 9 holding registers, each of which are two bytes wide 
	// (according to the Modbus protocol).
	// 
	// The first holding register is read-only.  The other 2 - 9 are 
	// writeable, but it is suggested that only holding registers 3 - 9 be 
	// written to when working with the SeaIO 46x modules.  SeaMAX provides
	// Ioctl library calls to alter registers 1 - 2 safely.
	//
	//
	// The layout of the holding registers is as follows (the field bit 
	// size is in parens):
	//
	//Reg 1: [----Marriage ID (8 bits)-----] [-----Module Type (8 bits)----]
	//
	//Reg 2: [-Baudrate (4)-] [-Parity (4)-] [-------Address (8 bits)------]
	//
	//               |--Port4--| |--Port3--|         |--Port2--| |--Port1--|
	//               |C   B   A| |C   B   A|         |C   B   A| |C   B   A|
	//Reg 3: [0] [0] [x] [x] [x] [x] [x] [x] [0] [0] [x] [x] [x] [x] [x] [x]
	//
	//       |-------Port B1 Presets-------| |-------Port A1 Presets-------|
	//Reg 4: [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x]
	//
	//       |-------Port A2 Presets-------| |-------Port C1 Presets-------|
	//Reg 5: [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x]
	//
	//       |-------Port C2 Presets-------| |-------Port B2 Presets-------|
	//Reg 6: [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x]
	//
	//       |-------Port B3 Presets-------| |-------Port A3 Presets-------|
	//Reg 7: [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x]
	//
	//       |-------Port A4 Presets-------| |-------Port C3 Presets-------|
	//Reg 8: [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x]
	//
	//       |-------Port C4 Presets-------| |-------Port B4 Presets-------|
	//Reg 9: [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x] [x]
	//
	type  = HOLDINGREG;
	start = 3;          //begin reading at register 3
	range = 7;          //read 7 holding registers, 2 bytes each.

	// When arranging the data, the high byte should come first followed
	// by the low byte.  A 1 is an input in the direction configuration
	// register and a 0 is an output.
	memset(&data, 0, sizeof(data));
	data[0] = 0x3F;     //We decide to make ports 7-12 inputs.
	data[1] = 0x00;     //Likewise ports 1-6 will be outputs.

	// Next we set the output presets, these will be the default values
	// at device power up.  Remember that a 1 indicates an output of roughly
	// 0V and a 0 indicates an output of roughly 5V.  Also remember that
	// this preset will only show up after power is removed from the device.
	data[2] = 0x22;    //This will be the configuration of port 2.
	data[3] = 0x11;    //This will be the configuration of port 1.
	data[4] = 0x44;    //This will be the configuration of port 4.
	data[5] = 0x33;    //This will be the configuration of port 3.
	data[6] = 0x66;    //This will be the configuration of port 6.
	data[7] = 0x55;    //This will be the configuration of port 5.
	
	// For this example it isn't necessary to configure ports 7-12, because
	// they are configured as inputs and will not have an output value.
	// NOTE: because we used memset to zero out the buffer, the default, if
	// ports 7-12 are changed to outputs, will be 0 for all those ports.

	// We can only write one register at a time,
	// to allow the module time to flash the register.
	for (i = 0; i < 7; i++)
	{
		// NOTE: The range value doesn't matter for a HOLDINGREG write
		result = myModule.Write(slaveId, type, start+i, 0, &data[2*i]);

		if (result < 0)
		{
			printf("ERROR: Set PIO config, Returned %d\n", result);
			return -1;
		}
	}
	PrintmoduleConfiguration(data);

	//Now let's read the module configuration.  It should be the one we
	//just wrote.  Unlike writing, you may read multiple registers with 
	//one call.
	memset(data, 0, sizeof(data));
	result = myModule.Read(slaveId, type, start, range, data);

	if (result < 0)
	{
		printf("ERROR: Reading PIO config, Returned %d\n", result);
		return -1;
	}
	else
	{
		PrintmoduleConfiguration(data);
	}

	// Close the connection and clean up the memory used by SeaMAX
	myModule.Close();

	// Return your comm port to it's original configuration.	
	if (comHandle > 0)  //only on RTU type
	{
		tcflush(comHandle, TCIFLUSH);  //clear the line
		tcsetattr(comHandle, TCSANOW, &original);  //set this config
	}

	return 0;
}

void PrintmoduleConfiguration(unsigned char *readHoldRegs)
{   
	printf("Current module Configuration ----------------\n");

	// HOLDING REGISTER 3

	// Now we print the default port I/O configuration (Identical to the
	// IOCTL_GET_PIO command using Ioctl()) 
	//
	//    0 = port is set to output mode
	//    1 = port is set to input mode
	//

	printf("\nCurrent Port Configuration (0 = output, 1 = input)\n");
	printf("Port A1: %d\n",(readHoldRegs[1] & 1));    
	printf("Port B1: %d\n",(readHoldRegs[1] >> 1) & 1); 
	printf("Port C1: %d\n",(readHoldRegs[1] >> 2) & 1);
	printf("Port A2: %d\n",(readHoldRegs[1] >> 3) & 1);
	printf("Port B2: %d\n",(readHoldRegs[1] >> 4) & 1);
	printf("Port C2: %d\n",(readHoldRegs[1] >> 5) & 1);
	printf("Port A3: %d\n",(readHoldRegs[0] & 1));
	printf("Port B3: %d\n",(readHoldRegs[0] >> 1) & 1);
	printf("Port C3: %d\n",(readHoldRegs[0] >> 2) & 1);
	printf("Port A4: %d\n",(readHoldRegs[0] >> 3) & 1);
	printf("Port B4: %d\n",(readHoldRegs[0] >> 4) & 1);
	printf("Port C4: %d\n",(readHoldRegs[0] >> 5) & 1);


	// The bit presets are divided by port (A1, B1, etc.) with one byte per
	// port. 
	//
	// Note: These presets are only valid for OUTPUTS (since you can't 
	// preset an input value) and will only appear the next time the module
	// is reset/looses power or changes configuration (input to output). 
	// It will NOT change the current settings.    

	printf("\nCurrent Port Presets (OFF = 0, ON = 1)\n");            

	printf("A1: 0x%2.2X\n",readHoldRegs[3]);
	printf("B1: 0x%2.2X\n",readHoldRegs[2]);
	printf("C1: 0x%2.2X\n",readHoldRegs[5]);
	printf("A2: 0x%2.2X\n",readHoldRegs[4]);
	printf("B2: 0x%2.2X\n",readHoldRegs[7]);
	printf("C2: 0x%2.2X\n",readHoldRegs[6]);
	printf("A3: 0x%2.2X\n",readHoldRegs[9]);
	printf("B3: 0x%2.2X\n",readHoldRegs[8]);
	printf("C3: 0x%2.2X\n",readHoldRegs[11]);
	printf("A4: 0x%2.2X\n",readHoldRegs[10]);
	printf("B4: 0x%2.2X\n",readHoldRegs[13]);
	printf("C4: 0x%2.2X\n",readHoldRegs[12]);

	return;
}
