/*
 * rad.c
 * SeaMAX for Linux Example Code
 *
 * This C code demonstrates the reading of SeaIO 470 analog inputs.
 *
 * Sealevel and SeaMAX are registered trademarks of Sealevel Systems
 * Incorporated.
 *
 *  2006-2017 Sealevel Systems, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>

// Inclusion of the SeaMAX library header is required
#include <seamaxlin.h>

int main(int argc, char * argv[])
{
	
	// Create the SeaMAX object for accessing SeaIO modules
	// NOTE the difference in the OOP class name (CSeaMaxLin) and the
	// nonOOP struct pointer (SeaMaxLin)
	SeaMaxLin *myModule = SeaMaxLinCreate();

	// A handle to the connection f.o.p.  (Only valid on RTU, not TCP)
	HANDLE comHandle;

	// Two termios structs, one to hold the current setup and one to
	// change so we can talk with our device.
	struct termios original, newtio;

	// Create the module string for use with SeaMAX Open(...)
	//
	// APPLICATION NOTE:
	//     
	//  "sealevel_rtu://dev/DEVNAME" is used for RS232, RS485, and USB 
	//  connections (DEVNAME is the special device file under /dev/)
	//
	//  "sealevel_tcp://xxx.xxx.xxx.xxx:yyy" is used for TCP/IP connections
	//  (where x's represent the ip address and y's the port number)
	//  By default your device will be DHCP enabled and use port 502.
	//  You may choose not to enter the port number and 502 will be used
	//  by default.  You may optionally enter your device name, if you have
	//  configured the device in this way.
	//
	char *portString = "sealevel_rtu://dev/ttyUSB0";

	// Each SeaIO module has a modbus compliant slave ID.  By default, 
	// modules are shipped with the hardware ID switched to zero, 
	// indicating a slave ID of 247.  Adjusting the hardware switch 
	// to any other address indicates that module will function at the 
	// switch's setting (i.e. a hardware switch setting of 2 indicates
	// that SeaIO module's slave ID is 2).  
	slave_address_t	slaveId = 1;

	// SeaMAX conforms to modbus addressing and function definitions.  
	// (Modbus addressing is one-indexed).  To read, for example, from 
	// inputs 3 to 5, use the following code: type = D_INPUTS, start = 3,
	// range = 3
	seaio_type_t	type;
	address_loc_t	start;
	address_range_t	range;
	adda_ext_config	extended_configuration;
	adda_config	configuration;
	
	// Input registers (according to modbus) are 2 bytes wide each
	unsigned char ad_input[2];
	unsigned int value;
	double percentage;
	char unit[10];

	// Open the SeaIO module by calling Open() from the SeaMAX library
	int result = SeaMaxLinOpen(myModule, portString);
	if (result != 0)
	{
		printf("ERROR: Open failed, Returned %d\n", result);
		return -1;
	}
	
	// Get the host computer communication parameters (only applicable on
	// RTU type connections (RS232, RS485, USB).  ie. Not for TCP type.
	// NOTE:  This is only necessary if you have configured your device to
	// use different communication parameters than the defaults.  Also note
	// that the comm parameters shown below are the defaults, if you have
	// used the IOCTLS to change the defaults, you should use a routine
	// something like this to be able to change your comm port, so you can
	// talk with your device.
	comHandle = SeaMaxLinGetCommHandle(myModule);

	if (comHandle > 0)  //only on RTU type
	{
		tcgetattr(comHandle, &original);  //save the current config
		bzero(&newtio, sizeof(newtio));   //clear the new config
	
		//9600 baud, 8 data bits, local use, readable
		newtio.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
		newtio.c_iflag = IGNPAR;  //ignore parity errors
		newtio.c_oflag = 0;       //raw data
		newtio.c_lflag = 0;       //enable canonical output
		newtio.c_cc[VTIME] = 10;  //1 second timeout on reads.

		tcflush(comHandle, TCIFLUSH);  //clear the line
		tcsetattr(comHandle, TCSANOW, &newtio);  //set this config
	}
	
	// Get the module configuration
	printf("---Getting ADDA module configuration data---\n");
	result = SeaMaxLinIoctl(myModule, slaveId, IOCTL_GET_ADDA_CONFIG, 
								&configuration);

	if (result != 0)
	{
		printf("ERROR: Get ADDA config, Returned %d\n", result);
		return -1;
	}

	// Determine the current A/D , D/A hardware configuration (jumper 
	// configuration)
	//
	// This Ioctl() call is computationally expensive and will alter 
	// the output values of the D/A channels.  This library call should be
	// used to determine the configuration of a SeaIO 470 module - however,
	// it does not have to be called for proper operation.  
	//
	// This call can be replaced with hard-coded values if the hardware 
	// configuration is known
	result = SeaMaxLinIoctl(myModule, slaveId, IOCTL_GET_ADDA_EXT_CONFIG, 
						&extended_configuration);

	if (result < 0)
	{
		printf("ERROR: Getting ext ADDA config, Returned %d\n", result);
		return -1;
	}

	printf("Done\n");

	// Set the configuration to be in +/- 10 volts mode reading in from 16
	// single ended inputs
	configuration.device.reference_offset = ANALOG_OFFSET;
	configuration.device.channel_mode = SINGLE_ENDED;
	configuration.channels.ch_1 = PLS_MIN_TEN;

	// Set the configuration back to the module - the module will retain
	// this configuration even after a power cycle
	result = SeaMaxLinIoctl(myModule, slaveId, IOCTL_SET_ADDA_CONFIG,
								&configuration);

	if (result < 0)
	{
		printf("ERROR: Setting A/D config, Returned %d\n", result);
		return -1;
	}

	printf("\nSet A/D to 16 single ended channels &");
	printf(" channel 1 range at +/- 10 Volts\n");

	// Read in a value from A/D channel 1
	type	= INPUTREG;
	start	= 1;
	range	= 1;

	memset(ad_input, 0, sizeof(ad_input));

	result = SeaMaxLinRead(myModule, slaveId, type, start, range, ad_input);

	if (result < 0)
	{
		printf("ERROR: Reading A/D, Returned %d\n", result);
		return -1;
	}

	// Convert from the modbus 2 bytes to an actual value
	value = (ad_input[0] << 8) + (ad_input[1]);

	if (configuration.channels.ch_1 == ZERO_TO_FIVE ||
				configuration.channels.ch_1 == ZERO_TO_TEN)
	{
		// We don't have to worry about sign conversion, so just display
		percentage = value / 4095.0;
	}
	else if (configuration.channels.ch_1 == PLS_MIN_FIVE || 
				configuration.channels.ch_1 == PLS_MIN_TEN)
	{
		// Adjust for the sign bit
		//
		// If the sign bit (0x800) is set and the channel mode is 
		// PLS_MIN_? then the value is negative
		// The easiest way to convert the value is to mask out the sign 
		// bit (value & 0x7FF) and then subtrace out the maximum value
		// (0x7FF or 2047).  The result will be a value between 0 and
		// -2047.0
		if (value & 0x800)
		{
			percentage = ((double)(value & 0x7FF)-2047.0) / 2047.0;
		}

		// If the sign bit was not set, then the value is positive and
		// the extra manipulation isn't necessary
		else
		{
			percentage = value / 2047.0;
		}

	}
	else
	{
		printf("ERROR: Unknown A/D config, Returned %d\n", result);
		return -1;
	}

	// If the hardware 10x gain jumper is enabled, the the maximum values
	// range from (0,0.5) or (0, 1)

	if (extended_configuration.ad_multiplier_enabled)
	{
		percentage = percentage / 10;
	}

	// In current loop mode, the two differential inputs are connected 
	// through a 249 Ohm precision resistor.  Since voltage = current *
	// resistance, the current is calculated as the voltage divided by the 
	// resistance.
	if (configuration.device.channel_mode == CURRENT_LOOP)
	{
		percentage = percentage / 249.0 * 1000.0;
		strcpy(unit, "mAmps");
	}
	else
	{
		strcpy(unit, "Volts");
	}

	switch(configuration.channels.ch_1)
	{
		case ZERO_TO_FIVE:	
		case PLS_MIN_FIVE:
		  printf("\nCH 1: %.2f %s\n", percentage * 5.0, unit);
	 	  break;
		case ZERO_TO_TEN:
		case PLS_MIN_TEN:
		  printf("\nCH 1: %.2f %s\n", percentage * 10.0, unit);
		  break;
	}

	// Close the connection and clean up the memory used by SeaMAX
	SeaMaxLinClose(myModule);
	SeaMaxLinDestroy(myModule);
	
	// Return your comm port to it's original configuration.	
	if (comHandle > 0)  //only on RTU type
	{
		tcflush(comHandle, TCIFLUSH);  //clear the line
		tcsetattr(comHandle, TCSANOW, &original);  //set this config
	}

	return 0;
}
