/*
 * wda.c
 * SeaMAX for Linux Example Code
 *
 * This C code demonstrates the writing of SeaIO 470
 * analog outputs.
 *
 * Sealevel and SeaMAX are registered trademarks of Sealevel Systems
 * Incorporated.
 *
 *  2006-2017 Sealevel Systems, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>

// Inclusion of the SeaMAX library header is required
#include <seamaxlin.h>

int main(int argc, char * argv[])
{
	// Create the SeaMAX object for accessing SeaIO modules
	// NOTE the difference in the OOP class name (CSeaMaxLin) and the
	// nonOOP struct pointer (SeaMaxLin)
	SeaMaxLin *myModule = SeaMaxLinCreate();

	// A handle to the connection f.o.p.  (Only valid on RTU, not TCP)
	HANDLE comHandle;

	// Two termios structs, one to hold the current setup and one to
	// change so we can talk with our device.
	struct termios original, newtio;

	// Create the module string for use with SeaMAX Open(...)
	//
	// APPLICATION NOTE:
	//     
	//  "sealevel_rtu://dev/DEVNAME" is used for RS232, RS485, and USB 
	//  connections (DEVNAME is the special device file under /dev/)
	//
	//  "sealevel_tcp://xxx.xxx.xxx.xxx:yyy" is used for TCP/IP connections
	//  (where x's represent the ip address and y's the port number)
	//  By default your device will be DHCP enabled and use port 502.
	//  You may choose not to enter the port number and 502 will be used
	//  by default.  You may optionally enter your device name, if you have
	//  configured the device in this way.
	//
	char *portString = "sealevel_rtu://dev/ttyUSB0";

	// Each SeaIO module has a modbus compliant slave ID.  By default, 
	// modules are shipped with the hardware ID switched to zero, 
	// indicating a slave ID of 247.  Adjusting the hardware switch 
	// to any other address indicates that module will function at the 
	// switch's setting (i.e. a hardware switch setting of 2 indicates
	// that SeaIO module's slave ID is 2).  
	slave_address_t	slaveId = 1;

	// SeaMAX conforms to modbus addressing and function definitions.  
	// (Modbus addressing is one-indexed).  To read, for example, from 
	// inputs 3 to 5, use the following code: type = D_INPUTS, start = 3,
	// range = 3
	seaio_type_t	type;
	address_loc_t	start;
	address_range_t	range;
	adda_ext_config	extended_configuration;
	
	// Holding registers (according to modbus) are 2 bytes wide each
	unsigned char da_output[2];
	unsigned short value;

	// Open the SeaIO module by calling Open() from the SeaMAX library
	int result = SeaMaxLinOpen(myModule, portString);
	if (result != 0)
	{
		printf("ERROR: Open failed, Returned %d\n", result);
		return -1;
	}
	
	// Get the host computer communication parameters (only applicable on
	// RTU type connections (RS232, RS485, USB).  ie. Not for TCP type.
	// NOTE:  This is only necessary if you have configured your device to
	// use different communication parameters than the defaults.  Also note
	// that the comm parameters shown below are the defaults, if you have
	// used the IOCTLS to change the defaults, you should use a routine
	// something like this to be able to change your comm port, so you can
	// talk with your device.
	comHandle = SeaMaxLinGetCommHandle(myModule);

	if (comHandle > 0)  //only on RTU type
	{
		tcgetattr(comHandle, &original);  //save the current config
		bzero(&newtio, sizeof(newtio));   //clear the new config
	
		//9600 baud, 8 data bits, local use, readable
		newtio.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
		newtio.c_iflag = IGNPAR;  //ignore parity errors
		newtio.c_oflag = 0;       //raw data
		newtio.c_lflag = 0;       //enable canonical output
		newtio.c_cc[VTIME] = 10;  //1 second timeout on reads.

		tcflush(comHandle, TCIFLUSH);  //clear the line
		tcsetattr(comHandle, TCSANOW, &newtio);  //set this config
	}

	// Determine the current A/D , D/A hardware configuration (jumper 
	// configuration)
	//
	// This Ioctl() call is computationally expensive and will alter the 
	// output values of the D/A channels.  This library call should be used
	// to determine the configuration of a SeaIO 470 on-board module's 
	// jumpers - however, it does not have to be called to use the features 
	// of the SeaIO 470.  
	//
	// This call can be replaced with hard-coded values if the hardware
	// configuration is known, however, since this is an example program
	// running on a unknown hardware configuration, we will use the Ioctl()
	// call to determine the on-board jumper configuration.
	printf("Getting the hardware jumper configuration ... ");
	result = SeaMaxLinIoctl(myModule, slaveId, IOCTL_GET_ADDA_EXT_CONFIG, 
						&extended_configuration);

	if (result < 0)
	{
		printf("ERROR: Getting hardware config, Returned %d\n", result);
		return -1;
	}

	printf("Done\n");

	// Clear D/A channel 1 (zero output voltage)
	type	= HOLDINGREG;
	start	= 1;
	range	= 1;

	// Zero channel1's output voltage
	memset(da_output, 0, sizeof(da_output));

	result = SeaMaxLinWrite(myModule,slaveId,type,start,range,da_output);

	if (result < 0)
	{
		printf("ERROR: Writing D/A channel 1, Returned %d\n", result);
		return -1;
	}

	printf("\nChannel1 output voltage set to 0.00 Volts\n");

	// Set channel 1's output voltage to 3.33 Volts
	type	= HOLDINGREG;
	start	= 1;
	range	= 1;

	// 4095 (0xFFF) is the maximum value that can be written out and
	// translates to either 5.0 volts or 10.0 volts, depending on the 
	// hardware jumper configuration.
	if (extended_configuration.da_channel_1_range == ZERO_TO_FIVE)
	{
		value = (unsigned short)((3.333 / 5.0) * 4095.0);
	}
	else if (extended_configuration.da_channel_1_range == ZERO_TO_TEN)
	{
		value = (unsigned short)((3.333 / 10.0) * 4095.0);
	}
	else
	{
		printf("ERROR: Unknown extended configuration.\n");
		return -1;
	}

	// Convert our value into a unsigned char array in the modbus format
	da_output[0] = ((value & 0xFF00) >> 8);
	da_output[1] = value & 0x00FF;

	// Perform the write to channel 1
	result = SeaMaxLinWrite(myModule,slaveId,type,start,range,da_output);

	if (result < 0)
	{
		printf("ERROR: Writing value to D/A, Returned %d\n", result);
		return -1;
	}

	printf("\nWrote value %04X (%.3f V) to D/A channel 1\n", value, 3.333);

	// Close the connection and clean up the memory used by SeaMAX
	SeaMaxLinClose(myModule);
	SeaMaxLinDestroy(myModule);
	
	// Return your comm port to it's original configuration.	
	if (comHandle > 0)  //only on RTU type
	{
		tcflush(comHandle, TCIFLUSH);  //clear the line
		tcsetattr(comHandle, TCSANOW, &original);  //set this config
	}

	return 0;
}
