/*
 * seamaxlin.cpp
 * SeaMAX for Linux
 *
 * This code implements a simple C++ wrapper class for the C Application 
 * Programming Interface to enable  * access to Sealevel Systems IO products.
 *
 * Sealevel and SeaMAX are registered trademarks of Sealevel Systems
 * Incorporated.
 *
 * © 2008-2017 Sealevel Systems, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the Lesser GNU General Public License
 * as published by the Free Software Foundation; either version
 * 3 of the License, or (at your option) any later version.
 * LGPL v3
 */

#include "seamaxlin.h"

CSeaMaxLin::CSeaMaxLin(void)
{
	SeaMaxPointer = SeaMaxLinCreate();
}

CSeaMaxLin::~CSeaMaxLin(void)
{
	SeaMaxLinDestroy(SeaMaxPointer);
}

// ----------------------------------------------------------------------------
/// \ingroup group_seamax_oop
/// \brief Open a SeaIO module.
/// This function will attempt to open a specified SeaIO module.  The method of
/// specification is a string formatted like "sealevel_rtu://dev/[your_serial]"
/// or "sealevel_tcp://x.x.x.x:y" where x's represent ip address and y the port.
/// If you leave out the port, 502 will be used as the default.  You may also 
/// enter the device's DHCP name like: "sealevel_tcp://Samwise". For SeaDAC Lite
/// modules use sealevel_d2x://xxxx where xxxx=8112 or 8115 etc..
///
/// \param[in] *filename           Filename to open.
///
/// \return int            Error code.
/// \retval 0              Successfully opened device.
/// \retval -ENAMETOOLONG  Too many characters in filename string.
/// \retval -EINVAL        Unallocated pointer.
/// \retval -EBADF         Invalid filename.
/// \retval -EBUSY         Communications medium busy.
/// \retval -EXDEV         Unable to initialize communications.
/// \retval -ENOMEM        Low memory.
/// \retval -EPERM         Unable to retrieve current communication settings.
// ----------------------------------------------------------------------------
int CSeaMaxLin::Open(char *filename)
{
	return SeaMaxLinOpen(SeaMaxPointer, filename);
}

// ----------------------------------------------------------------------------
/// \ingroup group_seamax_oop
/// \brief Close a SeaIO module.
/// This function will attempt to close a previously opened seaMaxModule
/// and clear any local data.  It will also close the connection interface
/// (socket or tty) and clean up any local data.  You must call this when you
/// are finished with the module.
///
///
/// \return int   Error code.
/// \retval 0     Successfully closed module.
// ----------------------------------------------------------------------------
int CSeaMaxLin::Close(void)
{
	return SeaMaxLinClose(SeaMaxPointer);
}

// ----------------------------------------------------------------------------
/// \ingroup group_seamax_oop
/// \brief Read from a module.
/// This is a multifunction read that can read from a module previously opened.
/// Any SeaIO devices attached to that module can then be read from in one of 
/// the following ways.
///
/// \param[in] slaveId           Address of the device you wish to read.
/// \param[in] type              The read type to preform.
/// \param[in] starting_address  Where to start the read; MODBUS is base 1.
/// \param[in] range             How many consecutive addresses to read.
/// \param[out] *data            Pointer to data storage buffer.
///
/// \return int      Error code.
/// \retval >0       Number of bytes of data in the buffer.
/// \retval -EBADF   No module open or write error.
/// \retval -EINVAL  Null buffer or too much data in buffer.
/// \retval -ENOMEM  Low memory.
/// \retval -ENODEV  Didn't receive response.
/// \retval -EFAULT  MODBUS exception.  First byte of buffer contains exception.
// ----------------------------------------------------------------------------
int CSeaMaxLin::Read(slave_address_t slaveId, seaio_type_t type, 
			  address_loc_t starting_address, address_range_t range,
			  void *data)
{
	return SeaMaxLinRead(SeaMaxPointer, slaveId, type, starting_address,
				range, data);
}

// ----------------------------------------------------------------------------
/// \ingroup group_seamax_oop
/// \brief Read from a SeaDAC Lite module.
/// This is a multifunction read that can read from a module previously opened.
/// Any SeaIO devices attached to that module can then be read from in one of 
/// the following ways.
///
/// \param[out] *data            Pointer to data storage buffer.
/// \param[in] len               Length of data to read.
///
/// \return int      Error code.
/// \retval >0       Number of bytes of data in the buffer.
/// \retval -ERANGE  Data size too large.
/// \retval -EIO  	 Read I/O error.
// ----------------------------------------------------------------------------
int CSeaMaxLin::Read(unsigned char *data, int len)
{
	return SeaDacLinRead(SeaMaxPointer, data, len);
}

// ----------------------------------------------------------------------------
/// \ingroup group_seamax_oop
/// \brief Write to a module.
/// This is a multifunction write that can write from a module previously opened.
/// Any SeaIO devices attached to that module can then be written to in one of 
/// the following ways.
///
/// \param[in] slaveId           Address of the device you wish to write.
/// \param[in] type              The write type to preform.
/// \param[in] starting_address  Where to start the write; MODBUS is base 1.
/// \param[in] range             How many consecutive addresses to write.
/// \param[in] *data             Pointer to data buffer.
///
/// \return int      Error code.
/// \retval >0       Number of bytes of data in the buffer.
/// \retval -EBADF   No module open or write error.
/// \retval -EINVAL  Null buffer or too much data in buffer.
/// \retval -ENOMEM  Low memory.
/// \retval -ENODEV  Didn't receive response.
/// \retval -EFAULT  MODBUS exception.  First byte of buffer contains exception.
// ----------------------------------------------------------------------------
int CSeaMaxLin::Write(slave_address_t slaveId, seaio_type_t type, 
			   address_loc_t starting_address,
			   address_range_t range, unsigned char *data)
{
	return SeaMaxLinWrite(SeaMaxPointer, slaveId, type, starting_address,
				range, data);
}

// ----------------------------------------------------------------------------
/// \ingroup group_seamax_oop
/// \brief Write to a SeaDAC Lite module.
/// This is a multifunction write that can write to a module previously opened.
/// Any SeaIO devices attached to that module can then be written to in one of 
/// the following ways.
///
/// \param[out] *data            Pointer to data storage buffer.
/// \param[in] len               Length of data to read.
///
/// \return int      Error code.
/// \retval >0       Number of bytes of data in the buffer.
/// \retval -ERANGE  Data size too large.
/// \retval -EIO  	 Read I/O error.
// ----------------------------------------------------------------------------
int CSeaMaxLin::Write(unsigned char *data, int len)
{
	return SeaDacLinWrite(SeaMaxPointer, data, len);
}

/// \ingroup group_seamax_oop
/// \brief	Read the entire PIO space of a SeaDAC Lite module
///
/// \param[out]	data
///
/// \retval		0	Success.
/// \retval		-1	Invalid model number.
/// \retval		-2	Unknown connection type.
///
/// Reads the entire PIO space, inputs and outputs alike.
///
/// \warning	Sufficient data must be allocated for data before calling
///	this function.  For instance, the 8126 has 32 bits of PIO, therefore
/// data should be allocated with 4 bytes of data before calling this function.
///
/// \note	Only available for SeaDAC 8126.
int CSeaMaxLin::GetPIO(unsigned char* data)
{
	return SeaDacGetPIO(SeaMaxPointer, data);
}

/// \ingroup group_seamax_oop
/// \brief	Write the entire PIO space of a SeaDAC Lite module
///
/// \param[in]	data
///
/// \retval		0	Success.
/// \retval		-1	Invalid model number.
/// \retval		-2	Unknown connection type.
///
/// Writes the entire PIO space
///
/// \note	PIO space writes will not affect those pins marked as inputs.
/// \note	Only available for SeaDAC 8126.
int CSeaMaxLin::SetPIO(unsigned char* data)
{
	return SeaDacSetPIO(SeaMaxPointer, data);
}

/// \ingroup group_seamax_oop
/// \brief	Sets the PIO direction 
///
/// \param[in]	data
///
/// \retval		0	Success.
/// \retval		-1	Invalid model number.
/// \retval		-2	Unknown connection type.
///
/// \note	Only available for SeaDAC 8126.
int CSeaMaxLin::SetPIODirection(unsigned char* data)
{
	return SeaDacSetPIODirection(SeaMaxPointer, data);
}

/// \ingroup group_seamax_oop
/// \brief	Reads the PIO direction 
///
/// \param[in]	data
///
/// \retval		0	Success.
/// \retval		-1	Invalid model number.
/// \retval		-2	Unknown connection type.
///
/// \note	Only available for SeaDAC 8126.
int CSeaMaxLin::GetPIODirection(unsigned char* data)
{
	return SeaDacGetPIODirection(SeaMaxPointer, data);
}
	
// ----------------------------------------------------------------------------
/// \ingroup group_seamax_oop
/// \brief Multi-function tool to configure and access SeaIO specific features.
/// Ioctl can be used to get and set device parameters and states.
///
/// \param[in] slaveId           Address of the device you wish to access.
/// \param[in] which             Which type of Ioctl you wish to call.
/// \param     *data             Buffer to store or retrieve data from.
///
/// \return int     Error code.
/// \retval >0       Number of bytes of data in the buffer.
/// \retval -EBADF   No module open or write error.
/// \retval -EINVAL  Null buffer or too much data in buffer.
/// \retval -ENOMEM  Low memory.
/// \retval -ENODEV  Didn't receive response.
/// \retval -EFAULT  MODBUS exception.  First byte of buffer contains exception.
// ----------------------------------------------------------------------------
int CSeaMaxLin::Ioctl(slave_address_t slaveId, IOCTL_t which, void *data)
{
	return SeaMaxLinIoctl(SeaMaxPointer, slaveId, which, data);
}

/// \ingroup	group_seamax_oop
/// \brief	Sets the intermessage delay time.
///
/// \param[in]	delay	
///
/// \retval	>=0		Success.
/// \retval -EBADF   No open module.
/// \retval -ENODEV  No open connection.
/// \retval -EPERM   Delay must be > 1.
///
/// This function is used to set the intermessage delay time in effort to 
/// optimize the number of messages that can be sent.  This time must be at
/// least 4 character times long.  The default is 1ms.
///
/// The parameter 'delay' is the delay (in milliseconds) to wait between 
/// sending back to back messages.
// ----------------------------------------------------------------------------
int CSeaMaxLin::set_intermessage_delay(int delay)
{
	return SeaMaxLinSetIMDelay(SeaMaxPointer, delay);
}

// ----------------------------------------------------------------------------
/// \ingroup group_seamax_oop
/// \brief Get the handle to the actual communication medium.
/// For RTU types it will be a normal file pointer, but TCP will be a socket
/// handle.  For this reason, it was decided to only return RTU file pointers.
/// An attempt to get the socket handle will return an ENODEV error.
///
/// \return int      Error code.
/// \retval >0       Pointer to communications termios struct.
/// \retval -EBADF   No open module.
/// \retval -ENODEV  This is not an RTU type connection.
// ----------------------------------------------------------------------------
HANDLE CSeaMaxLin::getCommHandle(void)
{
	return SeaMaxLinGetCommHandle(SeaMaxPointer);
}

